/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*           _____________________________________________________              */
/*           | H Y T E C   8 4 0 1   A D C  A s y n  D r i v e r |              */
/*           -----------------------------------------------------              */
/*                                                                              */
/*  Source file name      :-      drvHy8401ip.c                                 */
/*                                                                              */
/*  Initial creation date :-      02-Jan-2009                                   */
/*                                                                              */
/*  Original Developers   :-      Graham Cross / Mark Woodward                  */
/*  Subsequent Developers :-      Jim Chen                                      */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :-                                                              */
/*                                                                              */
/*                                                                              */
/*  (C)2009 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    02/01/09    |  GSC/MAW  | 0.0     | Initial creation                      */    
/*                |           |         |                                       */
/*    26/05/09    |  JC/GSC   | 0.1     | Interfaces implementation             */    
/*                |           |         |                                       */
/*    26/06/09    |  JC       | 1.0     | Finished all functions except MCA and */    
/*                |           |         | EPID. First release.                  */
/*                |           |         |                                       */
/*    03/07/09    |  JC       | 1.1     | 1. Fixed online mode change bug       */ 
/*                |           |         | 2. Implemented MCA and EPID callback  */    
/*    18/07/09    |  JC       | 1.2     | 1. Seperated the single conversion    */ 
/*                |           |         |    interrupt from the number of       */    
/*                |           |         |    conversion interrupt. Also created */
/*                |           |         |    seperate queue task for singel     */
/*                |           |         |    conversion interrupt.              */
/*                |           |         | 2. Changed to automatic enable when   */
/*                |           |         |    it is set to continuous mode       */
/*    29/07/09    |  JC       | 1.3     | 1. Done a workaround to stop hardware */ 
/*                |           |         |    trigger state staying issue        */    
/*                |           |         | 2. Killed a bug for gated mode that   */
/*                |           |         |    starting ADC without looking at    */
/*                |           |         |    software inhibit                   */
/*                |           |         | 3. When any mode changes now, enable  */
/*                |           |         |    reenable settings need to be reset */
/*    31/07/09    |  JC       | 1.4     | Fixed a bug that normal ai and		*/ 
/*                |           |         |    waveform do not update in trigger  */    
/*                |           |         |    and gated mode                     */    
/*    04/08/09    |  JC       | 1.5     | 1. Fixed a bug that NORD value isn't  */ 
/*                |           |         |    set correctly.                     */    
/*                |           |         | 2. Re-arrange the waveform data store */    
/*                |           |         |    to start at the beginning of the   */    
/*                |           |         |    array instead of at the end.       */    
/*    06/08/09    |  JC       | 1.6     | Re-defined average overflow flag		*/ 
/*                |           |         |    conditions.                        */    
/*    07/08/09    |  JC       | 1.7     | Fixed bugs for trigger mode and		*/ 
/*                |           |         |    gated mode waveform record.        */    
/*    10/11/09    |  JC       | 1.8     | Fixed a bug in INTSR that should		*/ 
/*                |           |         |    detect which interrupt to route    */    
/*    12/11/09    |  JC       | 1.9     | Fixed a bug when using continuous		*/ 
/*                |           |         |    mode with fastADC, the ADC wouldn't*/    
/*                |           |         |    start.                             */    
/*    15/06/10    |  JC       | 2.0     | Added read csr after clearing			*/ 
/*                |           |         |    interrupt source in order to		*/    
/*                |           |         |    guarantee the write successful		*/    
/*    25/08/10    |  JC       | 2.1     | 1. Changed interrupt connection by 	*/ 
/*                |           |         |    using ipmIntConnect in the ipac 	*/    
/*                |           |         |    rather than using devLib since     */    
/*                |           |         |    devLib is not hardware architecture*/    
/*                |           |         |    independent. But this new code     */    
/*                |           |         |    needs the new ipac (new drvHy8002) */    
/*                |           |         |    which implements the ipmIntConnect.*/    
/*                |           |         | 2. Changed the parameter name of		*/    
/*                |           |         |    vmeSlotNum in Hy8401ipAsynInit to  */    
/*                |           |         |    carrierNum since it is the index   */    
/*                |           |         |    when adding a carrier              */    
/*    21/09/10    |  PL       | 2.2     | Changed (by Peter Leicester) for		*/ 
/*                |           |         |    clamping the maximum to about      */    
/*                |           |         |    +/-10v in Rescale() function       */    
/*    24/05/12    |  JC       | 2.3     | Added clearing rebooting flag in		*/ 
/*                |           |         |    init to avoid some VxWorks hanging */    
/*    28/06/13    |  JC       | 2.4     | Changed to rid bogus interrupt		*/ 
/*                |           |         |           							*/    
/*                |           |         |    		                            */    
/********************************************************************************/

/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* asyn includes first */
#include <asynDriver.h>
#include <asynInt32.h>
#include <asynInt32Array.h>
#include <asynFloat64.h>
#include <asynFloat64Array.h>
#include <asynUInt32Digital.h>
#include <asynDrvUser.h>
 
/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8401ipAsyn.h"

/**************************************************************/
/* A couple of key parameters defining the ADC memory layout. */
#define BANKSIZE    0x10000         /* number of samples/bank */
#define MAX_MESSAGES  100           /* maximum number of messages */
#define MAX_ADC_CHANNELS    8       /* number of ADC channels on 8401 card */

/* The nominal uncalibrated range: the following values will be mapped
 * to [-10,+10]V.  Note that 0x8000 should map to 0V.  It is also
 * important that RAW_MIN + RAW_MAX = 0x10000. */
#define RAW_MIN  0x000F
#define RAW_MAX  0xFFF1
#define RAW_ZERO 0x8000


/* The following values define the calibrated range of RVAL values
 * returned by this driver.  These correspond, in general, to a range
 * of [-10,+10]V mapped into a 32-bit range.
 *
 * To fill the 32 bit range we offset the unsigned uncalibrated range
 * by its zero point and then shift everything by 16 bits. */

#define CAL_MIN   ((long) ((RAW_MIN - RAW_ZERO) << 16))
#define CAL_MAX   ((long) ((RAW_MAX - RAW_ZERO) << 16))

/* Identifiers in PROM. */
#define VITA4       "VITA4 "
#define HYTECID_LO  0x0300
#define HYTECID_HI  0x0080
#define PROM_MODEL  0x8401
#define PROM_MODEL_ALT1  0x8401


/* define bits in the 8401 CSR */
#define CSR_HF      0x0001          /* Set when lower conversion memory filled. */
#define CSR_F       0x0002          /* Set when upper conversion memory filled. */     
#define CSR_CC      0x0004          /* Set when requested sample count captured. */
#define CSR_MIS     0x0008          /* Set when memory update is inhibited. */
#define CSR_MII     0x0010          /* Set if interrupt generated by inhibit (MIS). */
#define CSR_EII     0x0020          /* Set to enable MIS interrupt. */
/* Bit 0x0040 unused. */
#define CSR_1M      0x0080          /* 1 => 64K samples, 0 => 128K samples per chan. */
#define CSR_HE      0x0100          /* Enable interrupt on HF set. */
#define CSR_FE      0x0200          /* Enable interrupt on F set. */
#define CSR_EE      0x0400          /* Enable end of sample sequence interrupt. */
#define CSR_ET      0x0800          /* Allow external enable to control ADC. */
#define CSR_XC      0x1000          /* 0 => use internal clock, 1 => external clock. */
#define CSR_ST      0x2000          /* Set to trigger capture of programmed samples. */
#define CSR_EX      0x4000          /* 0 => continuous capture, 1 => on trigger. */
#define CSR_ARM     0x8000          /* Set to enable capture into waveform RAM. */

#define EXTCSR_TEN  0x0001          /* Triggered status */
#define EXTCSR_SMI  0x0002          /* Software inhibit control and status. Write 1 to inhibit memory writing; write 0 to allow. Read 1 means hardware inhibit; read 0 means hardware allows */
#define EXTCSR_IS   0x0004          /* Single conversion completed flag */
#define EXTCSR_ISE  0x0008          /* Single conversion completed interrupt enable */

#define ENABLEADC 1
#define DISABLEADC 0

#define INIT 0
#define RUNTIME 1

#define DEFAULTAVERAGE 1            /* default average for ai */
#define DEFAULTSAMPLE 1             /* default sample for waveform */
#define DEFAULTOFFSET 0             /* default sample for waveform */

/* driver support flags */
/* Bit0 - gate mode support */
/* Bit1 - trigger mode support */
/* Bit2 - trigger mode negative offsets support */
#define DRIVER_SUPPORT     0x0003   /* support gated mode, trigger mode but not support trigger mode with negative offset at develping time. */

#define WAITING_TRIGGER         0           /* waiting for trigger */
#define TRIGGERRED_COLLECTING   1           /* triggerred and collecting data */
#define FINISHED_AVAILABLE      2           /* finished and data available */

#define USHORT unsigned short


/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* ADC Register and PROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The memory block occupied by the ADC in 16-bit memory is divided
 * here into three areas: registers, PROM and calibration data (also in
 * the prom, but separated out for coding convenience). */


/* The five application registers are captured in this data structure. */

typedef struct
{
    USHORT status;                  /* 0x00 */
    USHORT pointer;                 /* 0x02 */
    USHORT numConversions;          /* 0x04 */
    USHORT clockRate;               /* 0x06 */
    USHORT vector;                  /* 0x08 */
    USHORT extStatus;               /* 0x0A */
    USHORT unused[2];               /* 0x0C - 0x0F */
    USHORT data[MAX_ADC_CHANNELS];  /* 0x10 - 0x1F */
} REGISTERS;


/* This data structure captures an image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;            /* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */
    USHORT calType;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
    USHORT unused2;                 /* 0x9C */
    USHORT WLO;                     /* 0x9E */
} ID_PROM;


/* Calibration data comes in one of two flavours: either three
 * breakpoints per channel or five breakpoints per channel, depending on
 * the version of the ADC. */

typedef union
{
    struct CAL3PT
    {
        USHORT ngFS;
        USHORT zero;
        USHORT psFS;
    } Cal3pt[MAX_ADC_CHANNELS];

    struct CAL5PT
    {
        USHORT ngFS;
        USHORT ngHS;
        USHORT zero;
        USHORT psHS;
        USHORT psFS;
    } Cal5pt[MAX_ADC_CHANNELS];
} CALIBRATION_DATA;

/* The application registers, ID PROM and calibration data are laid out
 * as a contiguous block. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x1F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x20 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
    CALIBRATION_DATA cal;                       /* 0xA0 -      */
} IO_MEMORY;


/* set to non-zero to output verbose messages */
static int      verbose = 0;

typedef enum{		/* enumerated type of the driver's operational modes */
	hy8401Continuous,
	hy8401Triggered,
	hy8401Gated,
} hy8401Modes;


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* Enumerate all available ASYN commands for this driver.
   ...always let the hy8401_MAX_COMMANDS be last! */
typedef enum{
    /* Write commands */
    hy8401SweepData = 0,
    hy8401Data,
	hy8401ScanPeriod,
	hy8401SetAverage,
	hy8401SetSample,
	hy8401SetMode,
	hy8401SetOffset,
	hy8401EnableState,
	hy8401SoftTrigger,
	hy8401ReEnableState,
	hy8401ClearBuffer,
	hy8401SetClockRate,
	hy8401SetExtClock,

    /* Read commands */
	hy8401AverageOverflow,
	hy8401Overflow,
	hy8401BufferCount,
	hy8401GateTriggerState,
	hy8401Support,
	hy8401_MAX_COMMANDS,		/* must always be last! */
} hy8401_cmd_enum;
	
typedef struct {
	hy8401_cmd_enum command;
	char *commandString;
} hy8401_command_t;

hy8401_command_t hy8401_commands[hy8401_MAX_COMMANDS] = {
	{hy8401SweepData,	        "DATA"},
	{hy8401Data,		        "GENDATA"},
	{hy8401ScanPeriod,		    "SCAN_PERIOD"},
	{hy8401SetAverage,		    "SETAVERAGE"},
	{hy8401SetSample,		    "SETSAMPLE"},
	{hy8401SetMode,		        "SETMODE"},
	{hy8401SetOffset,	        "SETOFFSET"},
	{hy8401EnableState,		    "ENABLE"},
	{hy8401SoftTrigger,			"SOFTTRIGGER"},
	{hy8401ReEnableState,		"REENABLE"},
	{hy8401ClearBuffer,			"CLEARBUFFER"},
	{hy8401SetClockRate,		"SETCLOCKRATE"},
	{hy8401SetExtClock,			"SETEXTCLOCK"},

	{hy8401AverageOverflow,		"AVERAGEOVERFLOW"},
	{hy8401Overflow,			"OVERFLOW"},
	{hy8401BufferCount,			"BUFFERCOUNT"},
	{hy8401GateTriggerState,	"GATETRIGGERSTATE"},
	{hy8401Support,				"SUPPORT"},
};

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*  Asyn Data Structure	 -	internal driver structure						  */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct drvHy8401Pvt
{
	char * portName;                       /* the portname used for the IP card */

	volatile IO_MEMORY * adc;			   /* pointer to IP card A16 space registers, ID and calibration data */ 
	volatile USHORT * convMemoryBase;      /* pointer to IP card memory space */

    /* Saved calibration data read from IP card PROM  */
	/* ... remember the calibration type and the calibration parameters. */
    USHORT calType;
    struct CALDATA
    {
        USHORT ngFS;                        /* ADC output at -10V (NeGative Full Scale) */
        USHORT ngHS;                        /* ADC output at  -5V (NeGative Half Scale) */
        USHORT zero;                        /* ADC output at   0V (zero) */
        USHORT psHS;                        /* ADC output at  +5V (PoSitive Half Scale) */
        USHORT psFS;                        /* ADC output at +10V (PoSitive Full Scale) */
    } cal[MAX_ADC_CHANNELS];

    long ladcData[MAX_ADC_CHANNELS][BANKSIZE];/* ADC data for I/O Intr Int32Array */
    double dadcData[MAX_ADC_CHANNELS][BANKSIZE];/* ADC data for I/O Intr Float64Array */

	/* Storage for the current Configuration parameters */

    USHORT	carrierNum;		                /* carrier index number when adding carrier */
    USHORT	ipSlotNum;		                /* IP card slot number on IP carrier - boot time configuration only */
    USHORT	vectorNum;		                /* Interrupt vector number - boot time configuration only */
    int     clockRate;		                /* ADC internal sampling clock rate - boot and run time configuration possible */
    int		extClock;		                /* Selects internal or external clock - boot and run time configuration possible */
	hy8401Modes	mode;		                /* Select triggering modee - boot and run time configuration possible */
    int     samples;	                    /* Number of samples - boot and run time configuration possible */
	int		offset;			                /* Offset value (mode dependant) - boot and run time configuration possible */
	int		average;		                /* Number of readings to SUM for average - boot and run time configuration possible */
	int		fastAdc;		                /* flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records */
	epicsInt32 enableState;                 /* enable state */                  
	epicsInt32 reEnableState;               /* re-enable state */                  
    
    /* status definition */
    int     averageOverflow;                /* over flow flag. for gated and trigger mode only */    
    int     overflow;                       /* over flow flag. for gated and trigger mode only */    
    int     gateTriggerState;               /* gate/trigger mode state */

    int     lastPointer;                    /* for trigger/gated mode periodic waveform read */    

    /* not sure we are going to use these at this moment */
	int		aiType;			                /* Signal range settings - this is not supported on the 8401 */
	int		range;			                /* ADC range - this is not supported on the 8401 */
    int		firstChan;		                /* First channel to use - boot and run time configuration possible */
    int		lastChan;		                /* Last channel to use - boot and run time configuration possible */
	int		chanBuffSize;	                /* Channel buffer size - boot and run time configuration possible */
	int     chanData[ MAX_ADC_CHANNELS];    /* storage for the IP card last channel data values */
    
    int     sampleCounter;                  /* counter for continuous mode to count samples acquired in order to generate callbacks */

/* driver specific parameters */
    int		rebooting;		                /* Flag to indicate is the IOC is currently re-booting */

/* Work Queue Stuff */
    epicsTimerId timerId;
    epicsMutexId lock;
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       /* for report */
    int messagesFailed;                     /* for report */
    epicsMessageQueueId intSweepQId;

/* asyn bits */
	asynUser *localpasynUser;
    
	/* asyn interfaces to implement for this driver */
	asynInterface common;
	asynInterface int32;
	asynInterface int32Array;
	asynInterface float64;
	asynInterface float64Array;
	asynInterface uint32Digital;
	asynInterface drvUser;

	/* asyn interrupt (Inst I/O) interface to implement */
	void *int32InterruptPvt;
	void *float64InterruptPvt;
	void *int32ArrayInterruptPvt;
	void *float64ArrayInterruptPvt;
	void *uint32DigitalInterruptPvt;
} drvHy8401Pvt;

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*****************************************************************************/
/*                                                                           */
/*                      Miscellaneous Support Routines                       */
/*                                                                           */
/*****************************************************************************/

static void setupADC                    ( drvHy8401Pvt *pPvt, int runtime);
static void getcalibration              ( drvHy8401Pvt *pPvt, int verbose);
static int setClockRate                 ( drvHy8401Pvt *pPvt, int clockRate, int runtime  );
static int setExtClock                  ( drvHy8401Pvt *pPvt, int extClock, int runtime  );
static int softwareTrigger              ( drvHy8401Pvt *pPvt );
static int setSamples                   ( drvHy8401Pvt *pPvt, int runtime );
static int setHy8401Mode                ( drvHy8401Pvt *pPvt, int runtime  );
static int setHy8401EnableDisable       ( drvHy8401Pvt *pPvt, int enable );
static long Rescale                     ( int inLow, int inHigh, long outLow, long outHigh, double in );
static long calibrate                   ( drvHy8401Pvt *pPvt, double rawval, int channel );
static double toVoltage                 ( drvHy8401Pvt *pPvt, double rawval, int channel );
static int setOffset                    ( drvHy8401Pvt *pPvt, int runtime );
static void rebootCallback              ( void *drvPvt );
static void INTSR                       ( void *drvPvt ); /* Interrupt Service Routine */
static void intQueuedTask               ( drvHy8401Pvt *pPvt);
static void intSweepQueuedTask          (drvHy8401Pvt *pPvt);
static int checkAverageOverflow         ( drvHy8401Pvt *pPvt, int runtime );
/*********************************************************/
/*                                                      */
/* Asyn interface Structures and Function prototypes   */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
static void hy8401_Report                     (void *drvPvt, FILE *fp, int details);
static asynStatus hy8401_Connect              (void *drvPvt, asynUser *pasynUser);
static asynStatus hy8401_Disconnect           (void *drvPvt, asynUser *pasynUser);
static struct asynCommon hy8401_Common =      { hy8401_Report, hy8401_Connect, hy8401_Disconnect};

/* asynInt32 methods */
static asynStatus hy8401_int32Write           (void *drvPvt, asynUser *pasynUser, epicsInt32 value);
static asynStatus hy8401_int32Read            (void *drvPvt, asynUser *pasynUser, epicsInt32 *value);
static asynStatus hy8401_getBounds            (void *drvPvt, asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);
static asynInt32 hy8401_Int32 =               { hy8401_int32Write, hy8401_int32Read, hy8401_getBounds, NULL, NULL};
	
/* asynInt32Array methods */
static asynStatus hy8401_int32ArrayRead       (void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual);
static asynStatus hy8401_int32ArrayWrite      (void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t nelements);
static asynInt32Array hy8401_Int32Array =     { hy8401_int32ArrayWrite, hy8401_int32ArrayRead, NULL, NULL};

/* asynFloat64 methods */
static asynStatus hy8401_float64Write         (void *drvPvt, asynUser *pasynUser, epicsFloat64 value);
static asynStatus hy8401_float64Read          (void *drvPvt, asynUser *pasynUser, epicsFloat64 *value);
static asynFloat64 hy8401_Float64 =           { hy8401_float64Write, hy8401_float64Read, NULL, NULL};

/* asynFloat64Array methods */
static asynStatus hy8401_float64ArrayRead     (void *drvPvt, asynUser *pasynUser, epicsFloat64 *data, size_t nelements, size_t *nactual);
static asynStatus hy8401_float64ArrayWrite    (void *drvPvt, asynUser *pasynUser, epicsFloat64 *data, size_t nelements);
static asynFloat64Array hy8401_Float64Array = { hy8401_float64ArrayWrite, hy8401_float64ArrayRead, NULL, NULL};

/* asynUInt32Digital methods */
static asynStatus hy8401_uint32DigitalWrite   (void *drvPvt, asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask);
static asynStatus hy8401_uint32DigitalRead    (void *drvPvt, asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask);
static asynUInt32Digital hy8401_UInt32Digital = { hy8401_uint32DigitalWrite, hy8401_uint32DigitalRead, NULL, NULL, NULL, NULL, NULL};

/* asynDrvUser methods */
static asynStatus hy8401_drvUserCreate        (void *drvPvt, asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize);
static asynStatus hy8401_drvUserGetType       (void *drvPvt, asynUser *pasynUser, const char **pptypeName, size_t *psize);
static asynStatus hy8401_drvUserDestroy       (void *drvPvt, asynUser *pasynUser);
static asynDrvUser hy8401_DrvUser =           { hy8401_drvUserCreate, hy8401_drvUserGetType, hy8401_drvUserDestroy};

/*****************************************************************************/

/********************************************************************/
/*														           */
/* Local configurationfFunction definition for IOC shell function */
/*													             */
/****************************************************************/
/*
 * Where: -
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D),
 *    vectorNum = interrupt vector (0 = find one for me),
 *    intEnable = whether or not interrupts should be enabled,
 *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz.
 *     extClock = 0 = internal, 1 = extrenal,
 *         mode = 0 = continuous, 1 = triggered, 3 = gated,
 *      samples = number of samples (not applicable in gated modes),
 *       offset = Trigger mode: Offset from trigger point to first ADC reading to return
				  Gated mode:   Offset from the start (or end if negative) of the gated period to the ADC readings to return
				  Continuous:   Not applicable
 *       aiType = NOT SUPPORTED ON 8401  analogue input signal type  0 = diff, 1 = single-ended,
 *		  range = NOT SUPPORTED ON 8401  0 = +/- 5V, 1= +/1 10V
 *    firstChan = The first channel in use (0-7),
 *     lastChan = The last  channel in use (0-7),
 *	   chanBuff = MAx size of driver internal bufer for each channel.
 
*/
    
int Hy8401ipAsynInit(char *portName, int carrierNum, int ipSlotNum, int vectorNum,
					int samples, int average, int offset, int scanMode, int clockRate, 
                    int extClock, int fastADC)
{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok;
    int ishytec;
    int ismodel;
	int isAltmodel;

	asynStatus status = asynSuccess;
	drvHy8401Pvt *pPvt;

	/* allocate memory for local driver structure */
	pPvt = callocMustSucceed(1, sizeof(drvHy8401Pvt), "hy8401ipConfig");	
	
	pPvt->portName = epicsStrDup( portName );	                        /* store the asyn port name */
	pPvt->localpasynUser = pasynManager->createAsynUser(0, 0);          /* Create one asyn user to be used for debugging at startup */
	pasynTrace->setTraceMask( pPvt->localpasynUser, 0x11);              /*strictly speaking not correct use of asyn trace masks
																        however, it's the only way we can use asynPrint */
	/* Save the parameters in the driver structure    */
    pPvt->carrierNum		 = carrierNum;
    pPvt->ipSlotNum			 = ipSlotNum;
    pPvt->vectorNum          = vectorNum;
	pPvt->average			 = average;
    pPvt->samples			 = samples;
    pPvt->offset		     = offset;
	
    pPvt->clockRate			 = clockRate;                               /* Set MSB i.e. Add 0x8000; - If Running at 32 MHz  */
	pPvt->extClock		     = extClock;
    pPvt->mode				 = scanMode;
    pPvt->fastAdc			 = fastADC;
    pPvt->overflow           = 0;                                       /* no overflow */
    pPvt->gateTriggerState   = WAITING_TRIGGER;                         /* set to waiting for trigger */

    pPvt->enableState = 0;                                              /* disable first */
    pPvt->reEnableState = 0;                                            /* disable re-enable first */

/* Map Register address spaces  */
	pPvt->adc = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (pPvt->adc == NULL)
    {  
        /* can't register card register address space */
    	asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR,
		"configHy8401: Unable to register IO space\n");
  		errlogPrintf("configHy8401: Unable to register IO space\n");
        return asynError;
    }
/* map the memory address spaces  */
	pPvt->convMemoryBase = (USHORT *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrMem);
    if (pPvt->convMemoryBase == NULL)
    {  
        /* can't register card Memory address space */
    	asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR,
		"configHy8401: Unable to register memory space\n");
  		errlogPrintf("configHy8401: Unable to register memory space\n");
        return asynError;
    }

   /* Compare the HY8401 prom header, manufacturer's ID and card model
     * number with the expected values. */
    strok = memcmp((char *)&pPvt->adc->prom.vita4[0], VITA4, sizeof(pPvt->adc->prom.vita4)) == 0;
    ishytec = pPvt->adc->prom.hytecIdHigh == HYTECID_HI  &&
                  pPvt->adc->prom.hytecIdLow == HYTECID_LO;
    ismodel = pPvt->adc->prom.model == PROM_MODEL;
	isAltmodel = pPvt->adc->prom.model == PROM_MODEL_ALT1;

	 pPvt->calType = pPvt->adc->prom.calType; /* save the caliration type  */

	/* check if we have been asked to be more chatty */
  verbose=1;
  if (verbose)
    { 
        /* start with some basic information */
        printf("PROM: Contents at offset - %#08x\n", (int)&pPvt->adc->prom);
        printf("PROM: Prom header string - '%6s'\n", pPvt->adc->prom.vita4);
        printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
            (int)pPvt->adc->prom.hytecIdHigh, (int)pPvt->adc->prom.hytecIdLow, 
            ishytec ? hytecstr : "(Unknown)");
        printf("\n");
        /* output a bit more about this specific model and card */
        printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
            (int)pPvt->adc->prom.model, (int)pPvt->adc->prom.revision);
        printf("PROM: Serial number      - %hu\n", (int)pPvt->adc->prom.serialNumber);
        printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
            (int)pPvt->adc->prom.driverIdHigh, (int)pPvt->adc->prom.driverIdLow);
        printf("PROM: Device flags       - %#04hx\n", (int)pPvt->adc->prom.flags);
        printf("PROM: Bytes used         - %#04hx (%hd)\n",
            (int)pPvt->adc->prom.byteCount, (int)pPvt->adc->prom.byteCount);
        printf("\n");
    }
    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);
    if ((!ismodel) & (!isAltmodel))
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n",
            PROM_MODEL);
	}

/* Read and save the calibration data   */
	getcalibration(	pPvt, verbose);

/* Link with higher level routines */
	pPvt->common.interfaceType = asynCommonType;
	pPvt->common.pinterface  = (void *)&hy8401_Common;
	pPvt->common.drvPvt = pPvt;

	pPvt->int32.interfaceType = asynInt32Type;
	pPvt->int32.pinterface  = (void *)&hy8401_Int32;
	pPvt->int32.drvPvt = pPvt;

	pPvt->int32Array.interfaceType = asynInt32ArrayType;
	pPvt->int32Array.pinterface  = (void *)&hy8401_Int32Array;
	pPvt->int32Array.drvPvt = pPvt;

	pPvt->float64.interfaceType = asynFloat64Type;
	pPvt->float64.pinterface  = (void *)&hy8401_Float64;
	pPvt->float64.drvPvt = pPvt;

	pPvt->float64Array.interfaceType = asynFloat64ArrayType;
	pPvt->float64Array.pinterface  = (void *)&hy8401_Float64Array;
	pPvt->float64Array.drvPvt = pPvt;

	pPvt->uint32Digital.interfaceType = asynUInt32DigitalType;
	pPvt->uint32Digital.pinterface = (void *)&hy8401_UInt32Digital;
	pPvt->uint32Digital.drvPvt = pPvt;

	pPvt->drvUser.interfaceType = asynDrvUserType;
	pPvt->drvUser.pinterface = (void *)&hy8401_DrvUser;
	pPvt->drvUser.drvPvt = pPvt;
	
	/* registerport */
	status = pasynManager->registerPort(pPvt->portName,
										ASYN_MULTIDEVICE,
										1, /* autoconnect */
										0, /* medium priority */
										0); /* default stacksize */
	if (status != asynSuccess)
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig: ERROR: Can't register myself (port: %s)\n", pPvt->portName);
		return asynError;
	}

	/* register the interfaces with asyn */
    /* Interface Common */
	status = pasynManager->registerInterface(pPvt->portName, &pPvt->common);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig :  ERROR:Can't register common. (port: %s)\n",pPvt->portName);
		return asynError;
	}

	/* Interface int32 */
	status = pasynInt32Base->initialize(pPvt->portName, &pPvt->int32);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig:  ERROR:Can't register int32. (port: %s)\n",pPvt->portName);
		return asynError;
	}

	/* Interface int32Array */
	status = pasynInt32ArrayBase->initialize(pPvt->portName, &pPvt->int32Array);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig: ERROR: Can't register int32Array. (port: %s)\n",pPvt->portName);
		return asynError;
	}

	/* Interface float64 */
	status = pasynFloat64Base->initialize(pPvt->portName, &pPvt->float64);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig: Can't register float64.\n");
		return asynError;
	}

	/* Interface float64Array */
	status = pasynFloat64ArrayBase->initialize(pPvt->portName, &pPvt->float64Array);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig: Can't register float64Array.\n");
		return asynError;
	}

	/* Interface uint32Digital */
	status = pasynUInt32DigitalBase->initialize(pPvt->portName, &pPvt->uint32Digital);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig: ERROR: Can't register uint32Digital.\n");
		return asynError;
	}

	/* register interrupt source here (not all types will probably be needed for 8401) */
	pasynManager->registerInterruptSource(	pPvt->portName, &pPvt->int32,
											&pPvt->int32InterruptPvt);
	pasynManager->registerInterruptSource(	pPvt->portName, &pPvt->int32Array,
											&pPvt->int32ArrayInterruptPvt);
	pasynManager->registerInterruptSource(	pPvt->portName, &pPvt->float64,
											&pPvt->float64InterruptPvt);
	pasynManager->registerInterruptSource(	pPvt->portName, &pPvt->float64Array,
											&pPvt->float64ArrayInterruptPvt);
	pasynManager->registerInterruptSource(	pPvt->portName, &pPvt->uint32Digital,
											&pPvt->uint32DigitalInterruptPvt);

	/* Register pasynManager interface */
	status = pasynManager->registerInterface(pPvt->portName, &pPvt->drvUser);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig ERROR: Can't register drvUser\n");
		return asynError;
	}
	
	/* Connect pasynUser to device for debugging */
	status = pasynManager->connectDevice(pPvt->localpasynUser, pPvt->portName, 0);
	if (status != asynSuccess) 
	{
		asynPrint( pPvt->localpasynUser, ASYN_TRACE_ERROR, "### hy8401ipConfig, ERROR: connectDevice failed for scaler-sim\n");
		return asynError;
	}
	
    /* clear asyn message sent counters */
    pPvt->messagesSent = 0;
    pPvt->messagesFailed = 0;

/* set up interrupt, work queue and mutex stuff */
    pPvt->adc->reg.vector = vectorNum;            /* writes vector register offset 0x08 */
		if (ipmIntConnect(carrierNum, ipSlotNum, vectorNum, (void *)INTSR, (int)pPvt) < 0) {
      errlogPrintf("initIp330: failure connecting to interrupt\n");
      return asynError;
    }
    ipmIrqCmd(pPvt->carrierNum, pPvt->ipSlotNum, 0, ipac_irqEnable);


    pPvt->lock = epicsMutexCreate();
    pPvt->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8401intQueuedTask",
                           epicsThreadPriorityLow,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intQueuedTask,
                           pPvt) == NULL)
       errlogPrintf("Hy8401intQueuedTask epicsThreadCreate failure\n");
    
    pPvt->intSweepQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8401intSweepQTask",
                           epicsThreadPriorityHigh,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intSweepQueuedTask,
                           pPvt) == NULL)
       errlogPrintf("Hy8401intSweepQueuedTask epicsThreadCreate failure\n");

	pPvt->rebooting = 0;

    epicsAtExit(rebootCallback, pPvt);


/*******************************************************************/
/************* Working out real stuffs for ADC *********************/
/*******************************************************************/

    setupADC( pPvt, INIT );
	return asynSuccess;
}


int Hy8401ipAsynConfig(const char *portName, int aiType, int range, int firstChan, int lastChan, int chanBuffSize )
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
	drvHy8401Pvt *pPvt;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8401IpConfig, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8401IpConfig cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pPvt = pasynInterface->drvPvt;

    pPvt->aiType			 = aiType;
    pPvt->range				 = range;
    pPvt->firstChan          = firstChan;
    pPvt->lastChan           = lastChan;
    pPvt->chanBuffSize       = chanBuffSize;
    return(asynSuccess);
}



/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
static void hy8401_Report(void *drvPvt, FILE *fp, int details)
{
	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt;
    interruptNode *pnode;
    ELLLIST *pclientList;

    fprintf(fp, "Port: %s, carrier slot %d slot %d, base address=%p\n", 
            pPvt->portName, pPvt->carrierNum, pPvt->ipSlotNum, (void *)pPvt->adc);
    if (details >= 1) 
    {
        fprintf(fp, "    messages sent OK=%d; send failed (queue full)=%d\n",
                pPvt->messagesSent, pPvt->messagesFailed);
        
        /* Report int32 interrupts */
        pasynManager->interruptStart(pPvt->int32InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {
            asynInt32Interrupt *pint32Interrupt = pnode->drvPvt;
            fprintf(fp, "    int32 callback client address=%p, addr=%d, reason=%d\n",
                    (void *)pint32Interrupt->callback, pint32Interrupt->addr, 
                    pint32Interrupt->pasynUser->reason);
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->int32InterruptPvt);

        /* Report float64 interrupts */
        pasynManager->interruptStart(pPvt->float64InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {
            asynFloat64Interrupt *pfloat64Interrupt = pnode->drvPvt;
            fprintf(fp, "    float64 callback client address=%p, addr=%d, reason=%d\n",
                    pfloat64Interrupt->callback, pfloat64Interrupt->addr, 
                    pfloat64Interrupt->pasynUser->reason);
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->float64InterruptPvt);

        /* Report int32Array interrupts */
        pasynManager->interruptStart(pPvt->int32ArrayInterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {
            asynInt32ArrayInterrupt *pint32ArrayInterrupt = pnode->drvPvt;
            fprintf(fp, "    int32Array callback client address=%p, reason=%d\n",
                    pint32ArrayInterrupt->callback, 
                    pint32ArrayInterrupt->pasynUser->reason); 
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->int32ArrayInterruptPvt);

        /* Report float64Array interrupts */
        pasynManager->interruptStart(pPvt->float64ArrayInterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {
            asynFloat64ArrayInterrupt *pfloat64ArrayInterrupt = pnode->drvPvt;
            fprintf(fp, "    float64Array callback client address=%p, addr=%d, reason=%d\n",
                    pfloat64ArrayInterrupt->callback, pfloat64ArrayInterrupt->addr, 
                    pfloat64ArrayInterrupt->pasynUser->reason);
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->float64ArrayInterruptPvt);

    }
}

static asynStatus hy8401_Connect(void *drvPvt, asynUser *pasynUser)
{
	int signal;
	pasynManager->getAddr(pasynUser, &signal);
	pasynManager->exceptionConnect(pasynUser);
	return(asynSuccess);
}

static asynStatus hy8401_Disconnect(void *drvPvt, asynUser *pasynUser)
{
	pasynManager->exceptionDisconnect(pasynUser);
	return(asynSuccess);
}

/*****         asynInt32 methods       ******/
static asynStatus hy8401_int32Write(void *drvPvt, asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    ELLLIST *pclientList;
    interruptNode *pnode;
	int command = pasynUser->reason;
    int reason, period = 0;
    double power = 0.0;
    double dValue = 0.0;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt;
	
	pasynManager->getAddr(pasynUser, &channel);		/* Get the channel number (a pasynManager command)*/
	
//	printf("\n*** hy8401_int32Write *** command = %d (%s), channel = %d val = %i\n\n",
//				command, hy8401_commands[command].commandString, channel, value);

	if (channel < 0 || channel >= MAX_ADC_CHANNELS)			    /* check if it is a valid channel number */
	{
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, 
					"drvhy8401ip::hy8401_int32Write ERROR: feature index (channel = %d) invalid\n", channel);
		status = asynError;
		return status;
	}
	
	if (command < 0 || command >= hy8401_MAX_COMMANDS)	        /* check if the command exists */
	{
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, 
					"drvhy8401ip::hy8401_int32Write ERROR: feature index (channel = %d) invalid\n", channel);
		status = asynError;
		return status;
	}
	
	/* decide what to do depending on which command we received */
	switch(command)
	{
		case hy8401EnableState:
			setHy8401EnableDisable( pPvt, value );
			break;

		case hy8401ReEnableState:
			pPvt->reEnableState = value;
			break;

		case hy8401SetAverage: 
			pPvt->average = value;                                       /* user should be aware of that changing this will cause temporary turbulence of the measurement */
            checkAverageOverflow( pPvt, RUNTIME );
			break;
			
		case hy8401SetSample: 
			pPvt->samples = value;
            setSamples( pPvt, RUNTIME );                                 /* this is online command */
			break;
			
		case hy8401SetOffset: 
			pPvt->offset = value;                                        /* user should be aware of that changing this will cause temporary turbulence of the measurement */
            setOffset( pPvt, RUNTIME );
			break;
			
		case hy8401SetClockRate: 
			pPvt->clockRate = value;
            setClockRate( pPvt, pPvt->clockRate, RUNTIME );
			break;
			
		case hy8401SetExtClock: 
			pPvt->extClock = value;
            setExtClock( pPvt, pPvt->extClock, RUNTIME );
			break;
			
		case hy8401SetMode: 
			pPvt->mode = value;
            setupADC( pPvt, RUNTIME );
			break;
			
		case hy8401SoftTrigger: 
            status = softwareTrigger( pPvt );
			break;
			
		default:
        	asynPrint(	pasynUser, ASYN_TRACE_ERROR, 
						"drvhy8401ip::hy8401_int32Write command %s (%d) not implemented as int32-write!\n",
						hy8401_commands[command].commandString, command);
			status = asynError;
            return status;
			break;
	}

    /* notify the asyn user for setting changes */
    if ((command == hy8401SetClockRate) ||
        (command == hy8401SetExtClock) ||
        (command == hy8401SetMode) ||
        (command == hy8401SetSample) ||
        (command == hy8401SetOffset) ||
        (command == hy8401SetAverage))
    {
        /* Pass int32 interrupts */
        pasynManager->interruptStart(pPvt->int32InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynInt32Interrupt *pint32Interrupt = pnode->drvPvt;
            channel = pint32Interrupt->addr;
            reason = pint32Interrupt->pasynUser->reason;
            if ((reason == hy8401SetClockRate) ||
                (reason == hy8401SetExtClock) ||
                (reason == hy8401SetMode) ||
                (reason == hy8401SetSample) ||
                (reason == hy8401SetOffset) ||
                (reason == hy8401SetAverage))
            { 
                if((reason == hy8401SetSample) && (pPvt->samples > BANKSIZE)) reason = asynError; /* samples setting exceeds limit */
                if((reason == hy8401SetOffset) && (pPvt->average > BANKSIZE)) reason = asynError; /* average setting exceeds limit */
                pint32Interrupt->callback(pint32Interrupt->userPvt, 
                    pint32Interrupt->pasynUser, command);                          /* pass reason back to indicate the notification successful */
            }
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->int32InterruptPvt);
    }

    /* notify the asyn user for scan_period change */
    if (command == hy8401SetClockRate)
    {
        /* Pass float64 interrupts */
        pasynManager->interruptStart(pPvt->float64InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynFloat64Interrupt *pfloat64Interrupt = pnode->drvPvt;
            channel = pfloat64Interrupt->addr;
            reason = pfloat64Interrupt->pasynUser->reason;
            if (reason == hy8401ScanPeriod)
            { 
                /* This part is to figure out the relations in between 
                 * the setting to the real frequency. In 8401 case, 
                 * clockrate=0 -> 1Hz, 1 -> 2Hz, 2 - > 5Hz, 
                 *           3 -> 10Hz, 4 -> 20Hz, 5 -> 50Hz,
                 *           6 -> 100Hz, 7 -> 200Hz, 8 -> 500Hz,
                 *           9 -> 1kHz, 10 -> 2kHz, 11 -> 5kHz,
                 *           12 -> 10kHz, 13 -> 20kHz, 14 -> 50kHz,
                 *           15 -> 100kHz */
                period = (int) pPvt->clockRate % 3;
                if((period == 0) || (period == 2))
                    period = period * 2 + 1;
                else
                    period = 2;
                power = pow((double) 10, (double) ((int) (pPvt->clockRate / 3)));
                dValue = (double) 1 / (power * period);                             /* this convert the frequency to second */
                pfloat64Interrupt->callback(pfloat64Interrupt->userPvt, 
                    pfloat64Interrupt->pasynUser, dValue);                          /* pass scan period in second back to MCA and EPID users */
            }
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->float64InterruptPvt);
    }

	return( status );
}

static asynStatus hy8401_int32Read(void *drvPvt, asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel, localValue = 0;
    int i, pointer, numToAverage = 0, startPoint = 0, contAverage = 0;
    double dValue = 0.0;
	int command = pasynUser->reason;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt; 
    if (pPvt->rebooting) epicsThreadSuspendSelf();
	
	pasynManager->getAddr(pasynUser, &channel);		                                /* Get the channel number */
	
    switch(command)
    {
        case hy8401Data:
        case hy8401SweepData:
            pointer = pPvt->adc->reg.pointer;
 
            /* Note: average is already checked and corrected during the initialisation 
             * by checkAverageOverflow routine for continuous and trigger mode */
            switch(pPvt->mode)
            {
                case hy8401Continuous:
                    if (pPvt->average > pPvt->samples)
                    {
                        pPvt->averageOverflow = 1;                                  /* averaging overflow, this is an abnormal averaging */
                        contAverage = pPvt->samples;
                    }else
                    {
                        pPvt->averageOverflow = 0;                                  /* averaging normal */
                        contAverage = pPvt->average;
                    }
                    if (pointer >= contAverage) 
                        startPoint = pointer - contAverage;
                    else
                        startPoint = BANKSIZE - (contAverage - pointer);
                    numToAverage = contAverage;
                    break;
                case hy8401Triggered:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;   /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pointer <= pPvt->offset)                                    /* current pointer less than offset, take whatever is collected. ignore offset */
                    {
                        if (pointer < pPvt->average)                                /* current pointer also less than average */
                        {
                            startPoint = 0;                                         /* use all the data available */
                            numToAverage = pointer;
                        }else                                                       /* cuurent pointer greater than average */
                        {
                            startPoint = pointer - pPvt->average;                   /* use most recent data */
                            numToAverage = pPvt->average;
                        }   
                        pPvt->averageOverflow = 1;                                  /* averaging overflow, this is not normal averaging */
                    } else                                                          /* current pointer is greater than offset */
                    {
                        if ((pointer - pPvt->offset) >= pPvt->average)              /* if pointer minus offset is greater than average, take most recent data to average */
                        { 
                            startPoint = pointer - pPvt->average;
                            numToAverage = pPvt->average;
                            pPvt->averageOverflow = 0;                              /* This is the only normal averaging. */
                        }else                                                       /* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                        {
                            startPoint = pPvt->offset;
                            numToAverage = pointer - pPvt->offset;
                            pPvt->averageOverflow = 1;                              /* averaging overflow, this is an abnormal averaging */
                        }
                    }
                    break;
                case hy8401Gated:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;   /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pPvt->offset >= 0)                                           /* offset is possitive */
                    {
                        if (pointer < pPvt->offset)                                 /* current pointer less than offset, take whatever is collected. ignore offset */
                        {
                            if (pointer < pPvt->average)                            /* current pointer also less than average */
                            {
                                startPoint = 0;                                     /* use all the data available */
                                numToAverage = pointer;
                            }else                                                   /* cuurent pointer greater than average */
                            {
                                startPoint = pointer - pPvt->average;               /* use most recent data */
                                numToAverage = pPvt->average;
                            }   
                            pPvt->averageOverflow = 1;                              /* averaging overflow, this is not normal averaging */
                        } else                                                      /* current pointer is greater than offset */
                        {
                            if ((pointer - pPvt->offset) >= pPvt->average)          /* if pointer minus offset is greater than average, take most recent data to average */
                            { 
                                startPoint = pointer - pPvt->average;
                                numToAverage = pPvt->average;
                                pPvt->averageOverflow = 0;                          /* This is the only normal averaging */
                            }else                                                   /* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                            {
                                startPoint = pPvt->offset;
                                numToAverage = pointer - pPvt->offset;
                                pPvt->averageOverflow = 1;                          /* averaging overflow, this is an abnormal averaging */
                            }
                        }
                    }else                                                           /* offset is negative */
                    {
                        if (pointer >= (-pPvt->offset))
                        {
                            if (pPvt->average > (-pPvt->offset))
                            {
                                startPoint = pointer + pPvt->offset;
                                numToAverage = (-pPvt->offset);
                                pPvt->averageOverflow = 1;                          /* averaging overflow, abnormal averaging */
                            }else
                            {
                                startPoint = pointer - pPvt->average;               /* take whatever available */
                                numToAverage = pPvt->average;
                                pPvt->averageOverflow = 0;                          /* averaging normal */
                            }
                        }else 
                        {
                            if (pointer >= pPvt->average)
                            {
                                startPoint = pointer - pPvt->average;
                                numToAverage = pPvt->average;
            				    pPvt->averageOverflow = 0;                          /* averaging ok, the correct averaging */
                            }else
                            {
                                startPoint = 0;                                     /* take whatever available */
                                numToAverage = pointer;
                                pPvt->averageOverflow = 1;                          /* averaging overflow, abnormal averaging */
                            }
                        }
                    }
                    break;
                default:
                    status = asynError;
                    break;
            }
            for(i=0; i<numToAverage; i++)
            {
                dValue += (double) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel);
                startPoint++;
                if(startPoint >= BANKSIZE) startPoint = 0;                          /* if pointer reaches the bottom of the memory, reset it */
            }

            if (numToAverage != 0)
                localValue = (epicsInt32) (dValue / numToAverage);                  /* check deviding 0 for safety */
            else
                localValue = (epicsInt32) dValue; 
            *value = localValue;
/* printf("c=%d p=%d v=0x%x csr=0x%x extcsr=0x%x lastP=%d average=%d \n",channel, pPvt->adc->reg.pointer - 1, localValue, pPvt->adc->reg.status, pPvt->adc->reg.extStatus, pPvt->lastPointer, numToAverage); */                             /* for debug */
            break;
        case hy8401Support:
            *value = DRIVER_SUPPORT;                                                /* driver support */
            break;
        case hy8401Overflow:                                                        /* return overflow flag */
            *value = pPvt->overflow;
            break;
        case hy8401AverageOverflow:                                                 /* return average overflow flag */
            *value = pPvt->averageOverflow;
            break;
        case hy8401GateTriggerState:                                                /* return state for either trigger mode or gated mode */
            if (pPvt->mode == hy8401Continuous) 
                *value = asynError;                                                 /* continuous mode doesn't support this */
            else if (pPvt->mode == hy8401Triggered)
                localValue = pPvt->adc->reg.extStatus & EXTCSR_TEN;
            else if (pPvt->mode == hy8401Gated)
                localValue =  (~pPvt->adc->reg.status & CSR_MIS);

            if (localValue > 0)                                                     /* great than 0 means triggered */
                *value = TRIGGERRED_COLLECTING;
            else if (localValue == 0)                                               /* return whatever it is either waiting for trigger or finished or triigered by software */
                *value = pPvt->gateTriggerState;
            break;
        default:
            status = asynError;
	        *value = 0;
            break;
    }
	
	return( status );
}

static asynStatus hy8401_getBounds(void *drvPvt, asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = 65535;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}

/* asynInt32Array methods */
static asynStatus hy8401_int32ArrayRead(void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, pointer, startPoint = 0, numToReturn = 0;
	int command = pasynUser->reason;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt; 
    if (pPvt->rebooting) epicsThreadSuspendSelf();
	
	pasynManager->getAddr(pasynUser, &channel);		                              /* Get the channel number */
	
    switch(command)
    {
        case hy8401Data:
            pointer = pPvt->adc->reg.pointer;
            switch(pPvt->mode)
            {
                case hy8401Continuous:
                    if (pointer >= nelements) 
                        startPoint = pointer - nelements;
                    else
                        startPoint = BANKSIZE - (nelements - pointer);
                    numToReturn = nelements;
                    break;
                case hy8401Triggered:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;  /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pointer <= pPvt->offset) 
                    {
                        startPoint = 0;
                        numToReturn = pointer;
                    } else
                    {
                        startPoint = pPvt->offset;
                        numToReturn = pointer - pPvt->offset;
                    }
                    break;
                case hy8401Gated:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;  /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pPvt->offset >= 0)
                    {
                        if (pointer <= pPvt->offset) 
                        {
                            startPoint = 0;
                            numToReturn = pointer;
                        } else
                        {
                            startPoint = pPvt->offset;
                            numToReturn = pointer - pPvt->offset;
                        }
                    }
                    else
                    {
                        if (pointer < (-pPvt->offset)) 
                        {
                            startPoint = 0;
                            numToReturn = pointer;
                        } else
                        {
                            startPoint = pointer + pPvt->offset;
                            numToReturn = (-pPvt->offset);
                        }
                    }
                    break;
                default:
                    status = asynError;
                    break;
            }
            for(i=0; i<nelements; i++)
            {
                if (i < numToReturn)
                    *(data + i) = (epicsInt32) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel);
                else
                    *(data + i) = 0;
                startPoint++;
                if(startPoint >= BANKSIZE) startPoint = 0;                          /* if pointer reaches the bottom of the memory, reset it */
            }
            if (nelements > numToReturn)                                            /* if number to return less than number of elements, NORD = number of return */
                *nactual = numToReturn;
            else
                *nactual = nelements;                                               /* if number to return is greater or equal number of elements, NORD = number of elements */
            break;
        case hy8401SweepData:                                                       /* Please note, this returns 8 channel calibrated data of most recent conversion */
            pointer = pPvt->adc->reg.pointer - 1;
            if ((pPvt->mode == hy8401Triggered) || (pPvt->mode == hy8401Gated))
            {
               if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;        /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
            }

            for(i=0; i<MAX_ADC_CHANNELS; i++)
                *(data + i) = (epicsInt32) calibrate(pPvt, (double) (pPvt->convMemoryBase + i * BANKSIZE)[pointer], i);
 
            *nactual = MAX_ADC_CHANNELS;
            break;
        default:
            status = asynError;
            break;
    }

	status = asynSuccess;
	return( status );
}

static asynStatus hy8401_int32ArrayWrite(void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t nelements)
{
	asynStatus status = asynSuccess;
	return( status );
}

/* asynFloat64 methods */
static asynStatus hy8401_float64Write(void *drvPvt, asynUser *pasynUser, epicsFloat64 value)
{
	asynStatus status = asynSuccess;
	return( status );
}

static asynStatus hy8401_float64Read(void *drvPvt, asynUser *pasynUser, epicsFloat64 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int command = pasynUser->reason;
    int i, pointer, numToAverage = 0, startPoint = 0, contAverage = 0;
    double dValue = 0.0;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt; 
    if (pPvt->rebooting) epicsThreadSuspendSelf();
	
	pasynManager->getAddr(pasynUser, &channel);		/* Get the channel number */
	
    switch(command)
    {
        case hy8401Data:
            pointer = pPvt->adc->reg.pointer;
 
            /* Note: average is already checked and corrected during the initialisation 
             * by checkAverageOverflow routine for continuous and trigger mode */
            switch(pPvt->mode)
            {
                case hy8401Continuous:
                    if (pPvt->average > pPvt->samples)
                    {
                        pPvt->averageOverflow = 1;                                  /* averaging overflow, this is an abnormal averaging */
                        contAverage = pPvt->samples;
                    }else
                    {
                        pPvt->averageOverflow = 0;                                  /* averaging normal */
                        contAverage = pPvt->average;
                    }
                    if (pointer >= contAverage) 
                        startPoint = pointer - contAverage;
                    else
                        startPoint = BANKSIZE - (contAverage - pointer);
                    numToAverage = contAverage;
                    break;
                case hy8401Triggered:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;   /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pointer < pPvt->offset)                                     /* current pointer less than offset, take whatever is collected. ignore offset */
                    {
                        if (pointer < pPvt->average)                                /* current pointer also less than average */
                        {
                            startPoint = 0;                                         /* use all the data available */
                            numToAverage = pointer;
                        }else                                                       /* cuurent pointer greater than average */
                        {
                            startPoint = pointer - pPvt->average;                   /* use most recent data */
                            numToAverage = pPvt->average;
                        }   
                        pPvt->averageOverflow = 1;                                  /* averaging overflow, this is not normal averaging */
                    } else                                                          /* current pointer is greater than offset */
                    {
                        if ((pointer - pPvt->offset) >= pPvt->average)              /* if pointer minus offset is greater than average, take most recent data to average */
                        { 
                            startPoint = pointer - pPvt->average;
                            numToAverage = pPvt->average;
                            pPvt->averageOverflow = 0;                              /* This is the only normal averaging */
                        }else                                                       /* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                        {
                            startPoint = pPvt->offset;
                            numToAverage = pointer - pPvt->offset;
                            pPvt->averageOverflow = 1;                              /* averaging overflow, this is an abnormal averaging */
                        }
                    }
                    break;
                case hy8401Gated:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer;   /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pPvt->offset >= 0)                                           /* offset is possitive */
                    {
                        if (pointer < pPvt->offset)                                 /* current pointer less than offset, take whatever is collected. ignore offset */
                        {
                            if (pointer < pPvt->average)                            /* current pointer also less than average */
                            {
                                startPoint = 0;                                     /* use all the data available */
                                numToAverage = pointer;
                            }else                                                   /* cuurent pointer greater than average */
                            {
                                startPoint = pointer - pPvt->average;               /* use most recent data */
                                numToAverage = pPvt->average;
                            }   
                            pPvt->averageOverflow = 1;                              /* averaging overflow, this is not normal averaging */
                        } else                                                      /* current pointer is greater than offset */
                        {
                            if ((pointer - pPvt->offset) >= pPvt->average)          /* if pointer minus offset is greater than average, take most recent data to average */
                            { 
                                startPoint = pointer - pPvt->average;
                                numToAverage = pPvt->average;
                                pPvt->averageOverflow = 0;                          /* This is the only normal averaging */
                            }else                                                   /* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                            {
                                startPoint = pPvt->offset;
                                numToAverage = pointer - pPvt->offset;
                                pPvt->averageOverflow = 1;                          /* averaging overflow, this is an abnormal averaging */
                            }
                        }
                    }else                                                           /* offset is negative */
                    {
                        if (pointer >= (-pPvt->offset))
                        {
                            if (pPvt->average > (-pPvt->offset))
                            {
                                startPoint = pointer + pPvt->offset;
                                numToAverage = (-pPvt->offset);
                                pPvt->averageOverflow = 1;                          /* averaging overflow, abnormal averaging */
                            }else
                            {
                                startPoint = pointer - pPvt->average;               /* take whatever available */
                                numToAverage = pPvt->average;
                                pPvt->averageOverflow = 0;                          /* averaging normal */
                            }
                        }else 
                        {
                            if (pointer >= pPvt->average)
                            {
                                startPoint = pointer - pPvt->average;
                                numToAverage = pPvt->average;
            				    pPvt->averageOverflow = 0;                          /* averaging ok, the correct averaging */
                            }else
                            {
                                startPoint = 0;                                     /* take whatever available */
                                numToAverage = pointer;
                                pPvt->averageOverflow = 1;                          /* averaging overflow, abnormal averaging */
                            }
                        }
                    }
                    break;
                default:
                    status = asynError;
                    break;
            }

            for(i=0; i<numToAverage; i++)
            {
                dValue += toVoltage(pPvt, (double) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel), channel);
                startPoint++;
                if(startPoint > BANKSIZE) startPoint = 0;                           /* if pointer reaches the bottom of the memory, reset it */
            }
            if (numToAverage != 0)                                                  /* safety check */
                *value = dValue / numToAverage;
            else
                *value = dValue;
            break;
        case hy8401SweepData:                                                       /* for FastSweep record, the most recent acquired data is used */
            pointer = pPvt->adc->reg.pointer - 1;
            if ((pPvt->mode == hy8401Triggered) || (pPvt->mode == hy8401Gated))
            {
                if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer - 1;       /* for trigger/gated mode, use pointer stored by the interrupt only if adc address pointer is reset (=0) */
            }
 
            dValue = (double) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[pointer], channel);
            *value = dValue;
            break;
        default:
            status = asynError;
	        *value = 0;
            break;
    }
	
	return( status );
}

/* asynFloat64Array methods */
static asynStatus hy8401_float64ArrayRead(void *drvPvt, asynUser *pasynUser, epicsFloat64 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, pointer, startPoint = 0, numToReturn = 0;
	int command = pasynUser->reason;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt; 
    if (pPvt->rebooting) epicsThreadSuspendSelf();
	
	pasynManager->getAddr(pasynUser, &channel);		                              /* Get the channel number */
	
    switch(command)
    {
        case hy8401Data:
        case hy8401SweepData:
            pointer = pPvt->adc->reg.pointer;
            switch(pPvt->mode)
            {
                case hy8401Continuous:
                    if (pointer >= nelements) 
                        startPoint = pointer - nelements;
                    else
                        startPoint = BANKSIZE - (nelements - pointer);
                    numToReturn = nelements;
                    break;
                case hy8401Triggered:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer; /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pointer <= pPvt->offset) 
                    {
                        startPoint = 0;
                        numToReturn = pointer;
                    } else
                    {
                        startPoint = pPvt->offset;
                        numToReturn = pointer - pPvt->offset;
                    }
                    break;
                case hy8401Gated:
                    if (pPvt->adc->reg.pointer == 0) pointer = pPvt->lastPointer; /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                    if (pPvt->offset >= 0)
                    {
                        if (pointer <= pPvt->offset) 
                        {
                            startPoint = 0;
                            numToReturn = pointer;
                        } else
                        {
                            startPoint = pPvt->offset;
                            numToReturn = pointer - pPvt->offset;
                        }
                    }
                    else
                    {
                        if (pointer < (-pPvt->offset)) 
                        {
                            startPoint = 0;
                            numToReturn = pointer;
                        } else
                        {
                            startPoint = pointer + pPvt->offset;
                            numToReturn = (-pPvt->offset);
                        }
                    }
                    break;
                default:
                    status = asynError;
                    break;
            }
            for(i=0; i<nelements; i++)
            {
                if (i < numToReturn)
                    *(data + i) = toVoltage(pPvt, (epicsFloat64) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel), channel);
                else
                    *(data + i) = 0.0;
                startPoint++;
                if(startPoint >= BANKSIZE) startPoint = 0;                          /* if pointer reaches the bottom of the memory, reset it */
            }
            if (nelements > numToReturn)                                            /* if number to return less than number of elements, NORD = number of return */
                *nactual = numToReturn;
            else
                *nactual = nelements;                                               /* if number to return is greater or equal number of elements, NORD = number of elements */
            break;
        default:
            status = asynError;
            break;
    }

	status = asynSuccess;
	return( status );
}

static asynStatus hy8401_float64ArrayWrite(void *drvPvt, asynUser *pasynUser, epicsFloat64 *data, size_t nelements)
{
	asynStatus status = asynSuccess;
	return( status );
}

/****          Asyn uint32 digital routines      *******/
static asynStatus hy8401_uint32DigitalWrite(void *drvPvt, asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int command = pasynUser->reason;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt;
	
	pasynManager->getAddr(pasynUser, &channel);		/* Get the channel number (a pasynManager command)*/
	
	if (channel < 0 || channel >= MAX_ADC_CHANNELS)			    /* check if it is a valid channel number */
	{
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, 
					"drvhy8401ip::hy8401_int32Write ERROR: feature index (channel = %d) invalid\n", channel);
		status = asynError;
		return status;
	}
	
	if (command < 0 || command >= hy8401_MAX_COMMANDS)	        /* check if the command exists */
	{
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, 
					"drvhy8401ip::hy8401_uint32DigitalWrite ERROR: command = %d invalid\n", command);
		status = asynError;
		return status;
	}
	
	/* decide what to do depending on which command we received */
	switch(command)
	{
		case hy8401EnableState:
		case hy8401ReEnableState:
		case hy8401SetExtClock: 
		case hy8401SoftTrigger: 
			status = hy8401_int32Write(pPvt, pasynUser, value & mask);
			break;
					
		default:
        	asynPrint(	pasynUser, ASYN_TRACE_ERROR, 
						"drvhy8401ip::hy8401_uint32DigitalWrite command %s (%d) not implemented as unInt32Digital_write!\n",
						hy8401_commands[command].commandString, command);
			status = asynError;
			break;
	}
	return( status );
}

static asynStatus hy8401_uint32DigitalRead	(void *drvPvt, asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int command = pasynUser->reason;

	drvHy8401Pvt *pPvt = (drvHy8401Pvt*)drvPvt; 
    if (pPvt->rebooting) epicsThreadSuspendSelf();
	
	pasynManager->getAddr(pasynUser, &channel);		                 /* Get the channel number */
	
    switch(command)
    {
        case hy8401Support:
        case hy8401AverageOverflow:
        case hy8401Overflow:
        case hy8401GateTriggerState:
            status = hy8401_int32Read( drvPvt, pasynUser, (epicsInt32 *)value );
            *value &= mask;
            break;
        default:
            status = asynError;
	        *value = 0;
            break;
    }
	return( status );
}

/* asynDrvUser methods */
static asynStatus hy8401_drvUserCreate(void *drvPvt, asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize)
{
	int i;
	char *pstring;
	
	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvhy8401ip::drvUserCreate, attempting to create cmd: %s (hy8401_MAX_COMMANDS: %d)\n", drvInfo, hy8401_MAX_COMMANDS);
	
	for (i=0; i < hy8401_MAX_COMMANDS; i++)
	{
		pstring = hy8401_commands[i].commandString;
		if (epicsStrCaseCmp( drvInfo, pstring) == 0)
		{
			pasynUser->reason = hy8401_commands[i].command;
			if (pptypeName)
				*pptypeName = epicsStrDup(pstring);
			if (psize)
				*psize = sizeof( hy8401_commands[i].command);
			asynPrint(	pasynUser, ASYN_TRACE_FLOW,
						"drvhy8401ip::drvUserCreate, command created: %s\n", pstring);
			return( asynSuccess );
		}
	}
	asynPrint(	pasynUser, ASYN_TRACE_ERROR,
				"drvhy8401ip::drvUserCreate, unknown command: %s\n", drvInfo);
	return( asynError );			
}

static asynStatus hy8401_drvUserGetType(void *drvPvt, asynUser *pasynUser, const char **pptypeName, size_t *psize)
{
	int command = pasynUser->reason;
	*psize = 0;
	if ( pptypeName )
		*pptypeName = epicsStrDup( hy8401_commands[command].commandString);
	if ( psize )
		*psize = sizeof( command );
	return( asynSuccess );
}

static asynStatus hy8401_drvUserDestroy(void *drvPvt, asynUser *pasynUser)
{
	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvhy8401ip::drvUserDestroy. Function not implemented\n");
	return( asynSuccess );
}


static void rebootCallback(void *drvPvt)
{
    drvHy8401Pvt *pPvt = (drvHy8401Pvt *)drvPvt;
    pPvt->adc->reg.status &= ~CSR_ARM;                              /* Clear the ARM bit  */
    pPvt->rebooting = 1;				                            /* flag we are rebooting */
}

/*********************************************************/
/*                                                      */
/* Local Functions                                     */
/*                                                    */
/*****************************************************/

static void setupADC( drvHy8401Pvt *pPvt, int runtime)
{
/* this first write to the CSR sets this bit and clears all others */
    pPvt->adc->reg.status &= ~CSR_ARM;                                   /* stop ADC at the very beginning */
	pPvt->adc->reg.status = CSR_1M;
    if( pPvt->fastAdc )
        pPvt->adc->reg.extStatus = 0;                                    /* normal mode doesn't need interrupt on each conversion */
    else
    	pPvt->adc->reg.extStatus = EXTCSR_ISE;                           /* clear extension CSR and set single conversion interrupt enable for mca and EPID record support */

/* clear pointer to zero at start */
    pPvt->adc->reg.pointer = 0;

/* initialise average overflow flag */
	checkAverageOverflow( pPvt, runtime );	                             /* check average setting overflow */

/* set up the requested internal clock rate */
	setClockRate( pPvt, pPvt->clockRate, runtime );	                     /* writes clock rate register offset 0x06 */

/* set up the requested external / internal clock */
	setExtClock( pPvt, pPvt->extClock, runtime );	                     /* sets up the external clock bit state in the CSR */

/* set offset to check if it exceeds BANKSIZE when added to samples */
    setOffset( pPvt, runtime );

/* set up number of conversions */
	setSamples( pPvt, runtime );	                                      /* sets up number of conversion */

/* the mode setup of the ADC depends on the mode that has been requsted. */
	setHy8401Mode( pPvt, runtime );

/* disable the ADC at the beginning if it is not continuous mode and wait for the enable command */
    setHy8401EnableDisable( pPvt, pPvt->enableState );
}

static int setClockRate(drvHy8401Pvt *pPvt, int clockRate, int runtime  )
{
	if (pPvt->rebooting)  epicsThreadSuspendSelf();
    if ((clockRate < 0) || (clockRate > 15)) return( asynError );

    if( runtime )
    {
        pPvt->adc->reg.clockRate = clockRate;
    }
        pPvt->adc->reg.clockRate = clockRate;
    return(asynSuccess);
}

static int setExtClock(drvHy8401Pvt *pPvt, int extClock, int runtime )
{
	if (pPvt->rebooting)  epicsThreadSuspendSelf();
       
    if(runtime)                                                     /* change samples at realtime */
    { 
        pPvt->adc->reg.status &= ~CSR_ARM;                          /* disarm ADC first*/
        if (extClock)
	       pPvt->adc->reg.status |= CSR_XC;
	    else
	       pPvt->adc->reg.status &= ~CSR_XC;
        setHy8401EnableDisable( pPvt, pPvt->enableState );          /* start acquisition if it is enabled */
    }else
    {
        if (extClock)
	      pPvt->adc->reg.status |= CSR_XC;
	    else
	      pPvt->adc->reg.status &= ~CSR_XC;
    }    
    return( asynSuccess );
}

static int softwareTrigger(drvHy8401Pvt *pPvt )
{
	if (pPvt->rebooting)  epicsThreadSuspendSelf();
 
    if(pPvt->mode != hy8401Triggered ) 
        return asynError;                                           /* software trigger at realtime only applies to trigger mode */

    if(pPvt->enableState)                                           /* only trigger when enabled */
    {
        pPvt->adc->reg.status &= ~CSR_ST;                           /* clear software trigger latch */
        pPvt->adc->reg.status |= CSR_ST;                            /* software trigger to kick off the acquisition */
        pPvt->adc->reg.status &= ~CSR_ST;                           /* clear software latch to enable hardware trigger afterwards */
        pPvt->gateTriggerState = TRIGGERRED_COLLECTING; 
    }
    return( asynSuccess );
}

static int setSamples(drvHy8401Pvt *pPvt, int runtime )
{
	if (pPvt->rebooting)  epicsThreadSuspendSelf();
 
    if(runtime)                                                     /* change samples at realtime */
    { 
        switch(pPvt->mode)
        {
            case hy8401Continuous:
                pPvt->adc->reg.status &= ~CSR_ARM;                  /* disarm ADC first*/
                if(pPvt->samples > BANKSIZE) 
                {
                    pPvt->samples = BANKSIZE;
                    pPvt->overflow = 1;                             /* samples setting exceeds limit */
                } else
                    pPvt->overflow = 0;                             /* samples setting within limit */
                if( pPvt->fastAdc ) 
                    pPvt->adc->reg.numConversions = pPvt->samples;  /* put the number of conversion in */                    
                else
                    pPvt->sampleCounter = 0;                        /* clear sample counter */
                setHy8401EnableDisable( pPvt, pPvt->enableState );  /* start acquisition if it is enabled */
                break;
            case hy8401Triggered:
                pPvt->adc->reg.status &= ~CSR_ARM;                  /* disarm ADC first*/
                if(pPvt->samples + pPvt->offset > BANKSIZE) 
                {
                    pPvt->samples = BANKSIZE - pPvt->offset;
                    pPvt->overflow = 1;                             /* samples setting exceeds limit */
                } else
                    pPvt->overflow = 0;                             /* samples setting within limit */
                pPvt->adc->reg.numConversions = pPvt->samples + pPvt->offset;      /* put the new number of conversion in */

                if (pPvt->average <= pPvt->samples)                 /* if average is less than samples, fine */
                    pPvt->averageOverflow = 0;
                else                                                /* if average is greater than samples, averageoverflow */
                {
                    /*pPvt->average = pPvt->samples;*/                  /* only averaging the user interested data */
                    pPvt->averageOverflow = 1;                      /* averaging overflow, errorred */
                }

                setHy8401EnableDisable( pPvt, pPvt->enableState );  /* enable trigger if it is allowed */
                break;
            case hy8401Gated:                                       /* gated mode has nothing to do with the samples */
                break;
            default:
                return asynError;
                break;
        }    
    } else                                                          /* normal samples setting during the start up  */
    {
       switch(pPvt->mode)
       {
            case hy8401Continuous:
                if(pPvt->samples > BANKSIZE) 
                {
                    pPvt->samples = BANKSIZE;
                    pPvt->overflow = 1;                             /* samples setting exceeds limit */
                } else
                    pPvt->overflow = 0;                             /* samples setting within limit */

                if( pPvt->fastAdc ) 
                    pPvt->adc->reg.numConversions = pPvt->samples;  /* put the number of conversion in */                    
                else
                    pPvt->sampleCounter = 0;                        /* clear sample counter */
                break;
            case hy8401Triggered:
                if(pPvt->samples + pPvt->offset > BANKSIZE) 
                {
                    pPvt->samples = BANKSIZE - pPvt->offset;
                    pPvt->overflow = 1;                             /* samples setting exceeds limit */
                } else
                    pPvt->overflow = 0;                             /* samples setting within limit */

                if (pPvt->average <= pPvt->samples)                 /* if average is less than samples, fine */
                    pPvt->averageOverflow = 0;
                else                                                /* if average is greater than samples, averageoverflow */
                {
                    /*pPvt->average = pPvt->samples;*/                  /* only averaging the user interested data */
                    pPvt->averageOverflow = 1;                      /* averaging overflow, errorred */
                }

                pPvt->adc->reg.numConversions = pPvt->samples + pPvt->offset;
                break;
            case hy8401Gated:                                       /* gated mode has nothing to do with the samples */
                break;
            default:
                return asynError;
                break;
        }    
    }
    return( asynSuccess );
}

static int setHy8401Mode( drvHy8401Pvt *pPvt, int runtime )
{
	if (pPvt->rebooting) epicsThreadSuspendSelf();
    
    if ((pPvt->mode < hy8401Continuous) || (pPvt->mode > hy8401Gated)) return(asynError);

    if( runtime ) pPvt->adc->reg.status &= ~CSR_ARM;		        /* stop 8401 totally */

    switch(pPvt->mode)
    {
        case hy8401Continuous:
            pPvt->enableState = ENABLEADC;                          /* continuous mode enable adc at start */
            pPvt->sampleCounter = 0;                                /* clear sample counter */
            if( pPvt->fastAdc )                                     /* if it is normal mode, need conversion complete interrupt */
            {
                pPvt->adc->reg.status |= CSR_EX;                    /* enable trigger */
                pPvt->adc->reg.status |= CSR_EE;                    /* enable number of conversion interrupt */
            }
            break;
        case hy8401Triggered:                                       /* do nothing. The adc will be enabled by setEnable call */
            pPvt->enableState = DISABLEADC;                         /* disable adc at start for trigger mode */
            pPvt->reEnableState = DISABLEADC;                       /* disable re-enable at start for gated mode */
            pPvt->adc->reg.status |= CSR_EX;                        /* enable trigger */
            pPvt->adc->reg.status |= CSR_EE;                        /* enable number of conversion interrupt */
            break;
        case hy8401Gated:
            pPvt->enableState = DISABLEADC;                         /* disable adc at start for gated mode */
            pPvt->reEnableState = DISABLEADC;                       /* disable re-enable at start for gated mode */
            pPvt->adc->reg.extStatus |= EXTCSR_SMI;                 /* set software inhibit to preventing any writing to the memory */
            pPvt->adc->reg.status |= CSR_ET;                        /* enable memory inhibit bit */
            pPvt->adc->reg.status |= CSR_EII;                       /* enable inhibit interrupt */
            pPvt->adc->reg.status |= CSR_ARM;	                    /* Arm the 8401 to kick off acquiring continuously */
            break;
        default:
            return asynError;
            break;
    }
    return( asynSuccess );
}

static int setHy8401EnableDisable( drvHy8401Pvt *pPvt, int enable )
{
	if (pPvt->rebooting) epicsThreadSuspendSelf();
 
    pPvt->enableState = enable;                                     /* set the enable flag in pvt */
	if (enable)                                                     /* if enable */
    {
        switch (pPvt->mode)
        {
            case hy8401Continuous:
                pPvt->adc->reg.status |= CSR_ARM;	                /* Arm the 8401, enable acquiring */
                if( pPvt->fastAdc )                                 /* if it is normal mode, trigger the ADC */
                {
                    pPvt->adc->reg.status &= ~CSR_ST;               /* clear software trigger */
                    pPvt->adc->reg.status |= CSR_ST;                /* software trigger to kick off the acquisition */
                }
                break;
            case hy8401Triggered:
                pPvt->adc->reg.status &= ~CSR_EX;                   /* disable trigger to clear any pending triggers */
                pPvt->adc->reg.status |= CSR_EX;                    /* enable trigger */
                pPvt->adc->reg.status |= CSR_ARM;	                /* Arm the 8401, enable acquiring */
                pPvt->gateTriggerState = WAITING_TRIGGER;           /* status changed to waiting for next trigger */
                pPvt->adc->reg.pointer = 0;                         /* zero the address pointer */
                break;
            case hy8401Gated:
                pPvt->adc->reg.extStatus &= ~EXTCSR_SMI;            /* clear software iiinhibit to allow hardware inhibit */
                pPvt->gateTriggerState = WAITING_TRIGGER;           /* status changed to waiting for next trigger */
                pPvt->adc->reg.pointer = 0;                         /* zero the address pointer */
                break;
            default:
                return asynError;
                break;
        }
    }
	else                                                            /* if disable */
    {
        switch (pPvt->mode)
        {
            case hy8401Continuous:                                  /* do nothing. wait for the interrupt to finish the current acquisition then stop */
            case hy8401Triggered:
                pPvt->adc->reg.status &= ~CSR_ARM;	                /* disable 8401 */
                break;
            case hy8401Gated:
                 pPvt->adc->reg.extStatus |= EXTCSR_SMI;            /* set software inhibit to preventing any writing to the memory */
               break;
            default:
                return asynError;
                break;
        }
    }
	
   return( asynSuccess );
}

static int setOffset( drvHy8401Pvt *pPvt, int runtime )
{
    if(runtime)                                                     /* change samples at realtime */
    { 
        switch (pPvt->mode)
        { 
            case hy8401Triggered:
                pPvt->adc->reg.status &= ~CSR_ARM;	                /* disable 8401 */
                if(pPvt->samples + pPvt->offset > BANKSIZE) 
                {
                    pPvt->samples = BANKSIZE - pPvt->offset;
                    pPvt->overflow = 1;                             /* samples setting exceeds limit */
                } else
                    pPvt->overflow = 0;                             /* samples setting within limit */
                pPvt->adc->reg.numConversions = pPvt->samples + pPvt->offset;      /* put the new number of conversion in */
                setHy8401EnableDisable( pPvt, pPvt->enableState );  /* enable trigger if it is allowed */
                break;
            case hy8401Gated:
                if (pPvt->offset > BANKSIZE) 
                    pPvt->overflow = 1;
                else
                    pPvt->overflow = 0;
                break;
            default:
                return asynError;
                break;
        }
    } else
    {
        switch (pPvt->mode)
        { 
            case hy8401Triggered:
                 if(pPvt->samples + pPvt->offset > BANKSIZE) 
                {
                    pPvt->samples = BANKSIZE - pPvt->offset;
                    pPvt->overflow = 1;                             /* samples setting exceeds limit */
                } else
                    pPvt->overflow = 0;                             /* samples setting within limit */
                pPvt->adc->reg.numConversions = pPvt->samples + pPvt->offset;      /* put the new number of conversion in */
                break;
            case hy8401Gated:
                if (pPvt->offset > BANKSIZE) 
                    pPvt->overflow = 1;
                else
                    pPvt->overflow = 0;
                break;
            default:
                return asynError;
                break;
        }     
    }   
    return( asynSuccess );
}

static int checkAverageOverflow( drvHy8401Pvt *pPvt, int runtime )
{
    switch(pPvt->mode)
    {
        case hy8401Continuous:
            if (pPvt->average > BANKSIZE)
            {
                pPvt->averageOverflow = 1;
                pPvt->average = BANKSIZE;                           /* average greater than the memory size, trim it to the memory size */
            }
            else
                pPvt->averageOverflow = 0;
            break;
        case hy8401Triggered:
            if (pPvt->average <= pPvt->samples)                     /* if average is less than samples, fine */
                pPvt->averageOverflow = 0;
            else                                                    /* if average is greater than samples, averageoverflow */
            {
                /*pPvt->average = pPvt->samples;*/                      /* only averaging the user interested data */
                pPvt->averageOverflow = 1;                          /* averaging overflow, errorred */
            }
            break;
        case hy8401Gated:                                           /* gated mode has to be checked at runtime */
            if (pPvt->average > BANKSIZE)
            {
                pPvt->averageOverflow = 1;
                pPvt->average = BANKSIZE;                           /* average greater than the memory size, trim it to the memory size */
            }
            else
                pPvt->averageOverflow = 0;
            break;
        default:
            break;
    }    
    return( asynSuccess );
}


/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task Routines                   */
/*                                                    */
/*****************************************************/

static void INTSR(void *drvPvt)
{
    drvHy8401Pvt *pPvt = (drvHy8401Pvt *)drvPvt;
    int data[2], flag = 1, sweepFlag = 1;
    int sweepData[MAX_ADC_CHANNELS + 2];
    int i = 0;
    USHORT regData;

    data[0] = pPvt->adc->reg.pointer;                                               /* index 0 stores the address pointer */
    sweepData[0] = pPvt->adc->reg.pointer;                                          /* index 0 stores the address pointer */

	if ((pPvt->adc->reg.extStatus & EXTCSR_IS) && (pPvt->adc->reg.extStatus & EXTCSR_ISE)) 		/* if it is single conversion interrupt */
    {
       	pPvt->adc->reg.extStatus &= (~EXTCSR_IS);                                   /* clear singel conversion interrupt */
        regData = pPvt->adc->reg.extStatus;                                         /* to just ensure write successful */
        switch (pPvt->mode)
        {
            case hy8401Continuous:
            case hy8401Triggered:
	        	for (i=0; i<MAX_ADC_CHANNELS; i++)                                  /* the rest of the elements store 8 channel raw data */
    	        	sweepData[i + 2] = pPvt->adc->reg.data[i];
                break;
            case hy8401Gated:
				if ((!(pPvt->adc->reg.status & CSR_MIS)) 
                    && (!(pPvt->adc->reg.extStatus & EXTCSR_SMI)))	                /* only collect data when gate is open */
				{
	        		for (i=0; i<MAX_ADC_CHANNELS; i++)                              /* the rest of the elements store 8 channel raw data */
    	        		sweepData[i + 2] = pPvt->adc->reg.data[i];
				} else
					sweepFlag = 0;			      					                /* if gate not opened, don't wake up the queue */
                break;
            default:
                sweepFlag = 0;
                break;
        }

        /* Wake up aweep queue task which calls callback routines if it is the correct interrupt. Note, the mode setting guarantees the interrupt settings */
        if (sweepFlag)
        {
            if (epicsMessageQueueTrySend(pPvt->intSweepQId, sweepData, sizeof(sweepData)) == 0)
                pPvt->messagesSent++;
            else
                pPvt->messagesFailed++;
        }

        if( pPvt->mode == hy8401Continuous)                         /* only do this when it is continuous mode and slow ADC. fastADC wouldn't get here */
        {
            pPvt->sampleCounter++;
            if ( pPvt->sampleCounter >= pPvt->samples)
            {
                pPvt->sampleCounter = 0;
            
                /* Wake up queue task which calls callback routines */
                if (epicsMessageQueueTrySend(pPvt->intMsgQId, data, sizeof(data)) == 0)
                    pPvt->messagesSent++;
                else
                    pPvt->messagesFailed++;
            }
        }
    }
    else if(((pPvt->adc->reg.status & CSR_MII) & (pPvt->adc->reg.status & CSR_EII)) ||
    		((pPvt->adc->reg.status & CSR_HF) & (pPvt->adc->reg.status & CSR_HE)) ||
    		((pPvt->adc->reg.status & CSR_F) & (pPvt->adc->reg.status & CSR_FE)) ||
    		(pPvt->adc->reg.status & CSR_CC))										//only if there is an interreupt. conversion complete or gated interrupt
    {
        /* re-enable interrupt */
        switch (pPvt->mode)
        {
            case hy8401Continuous:                                  /* for mca mode (slow ADC), the code wouldn't get here */
                pPvt->adc->reg.status &= ~CSR_CC;                   /* clear interrupt */
                setHy8401EnableDisable( pPvt, pPvt->enableState );  /* only kicks off acquisition again when enabled. Otherwise acquisition stops from here */
                break;
            case hy8401Triggered:
                pPvt->adc->reg.status &= ~CSR_CC;                   /* clear interrupt */
                pPvt->gateTriggerState = FINISHED_AVAILABLE;        /* finished and data is available */
                break;
            case hy8401Gated:
                pPvt->adc->reg.status |= CSR_MII;                   /* clear interrupt */
                pPvt->gateTriggerState = FINISHED_AVAILABLE;        /* finished and data is available */
                break;
            default:
                flag = 0;
                break;
        }
        regData = pPvt->adc->reg.status;                            /* to just ensure write successful */

        /* Wake up queue task which calls callback routines if it is the correct interrupt. Note, the mode setting guarantees the interrupt settings */
        if (flag)
        {
            if (epicsMessageQueueTrySend(pPvt->intMsgQId, data, sizeof(data)) == 0)
                pPvt->messagesSent++;
            else
                pPvt->messagesFailed++;
        }
    }
}

static void intQueuedTask(drvHy8401Pvt *pPvt)
{
    int i;
    int reason;
    int data[2];
    ELLLIST *pclientList;
    interruptNode *pnode;
    epicsInt32 localValue;
    int pointer, startPoint = 0, STpoint = 0, channel, numOfElement = 0;
    double dValue = 0.0;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(pPvt->intMsgQId, data, sizeof(data));
                 
        pointer = data[0];

        numOfElement = pPvt->average;                                               /* assuming average setting is what we want */
        startPoint = pointer - 1;
        switch ( pPvt->mode )
        {
            case hy8401Continuous:                                                  /* continuous mode checks average overflow at runtime */
            case hy8401Triggered:                                                   /* trigger mode checks average overflow at runtime */
                if (pPvt->average > pPvt->samples)
                {
                    numOfElement = pPvt->samples;                                   /* use the data the user is interested */
                    pPvt->averageOverflow = 1;                                      /* averaging overflow, this is not normal averaging */
                }else
                    pPvt->averageOverflow = 0;                                      /* normal averaging */
                break;
            case hy8401Gated:                                                       /* gated mode checks average overflow at runtime */
                if (pPvt->offset >= 0)                                                  /* offset is possitive */
                {
                    if (pointer >= pPvt->offset)                                        /* please note the limitation of total samples of the gated mode is the BANKSIZE */
                    {
                        if (pointer - pPvt->offset < pPvt->average)                     /* if have enough data for averaging */
                        {
                            pPvt->averageOverflow = 1;                                  /* averaging overflow, errorred */
                            numOfElement = pointer - pPvt->offset;                      /* use only user interested data */
                        }
                            pPvt->averageOverflow = 0;                                  /* averaging normal */
                    }else
                    {
                        pPvt->averageOverflow = 1;                                      /* averaging overflow, errorred */
                        numOfElement = pointer;
                    }
                }else                                                                   /* offset is negative */
                {
                    if (pointer >= (-pPvt->offset))                                     /* collected data more than offset */
                    {
                        if (pPvt->average > (-pPvt->offset))                            /* but average is greater than offset */
                        {
                            pPvt->averageOverflow = 1;                                  /* averaging overflow, errorred */
                            numOfElement = (-pPvt->offset);                             /* only use data the user is interested */
                        }else
                            pPvt->averageOverflow = 0;                                  /* averaging normal */
                    }else                                                               /* collected data less than offset */
                    {
                        numOfElement = pointer;                                         /* take everything */
	   				pPvt->averageOverflow = 1;
                    }
                }
                break;
        }
        STpoint = startPoint;                                                       /* store start point */

        /* Pass int32 interrupts */
        pasynManager->interruptStart(pPvt->int32InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynInt32Interrupt *pint32Interrupt = pnode->drvPvt;
            channel = pint32Interrupt->addr;
            reason = pint32Interrupt->pasynUser->reason;
            startPoint = STpoint;                                                   /* re-gain teh start point */
            if (reason == hy8401Data)
            {

                dValue = 0.0;
                for(i=0; i<numOfElement; i++)
                {
                    dValue += (double) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel);
                    startPoint--;
                    if(startPoint < 0) startPoint = BANKSIZE - 1;                   /* if pointer reaches the bottom of the memory, reset it */
                }
                if (numOfElement != 0)                                              /* safety check */
                    localValue = (epicsInt32) (dValue / numOfElement);
                else
                    localValue = (epicsInt32) dValue; 

                pint32Interrupt->callback(pint32Interrupt->userPvt, 
                    pint32Interrupt->pasynUser, localValue);
            }

            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->int32InterruptPvt);

        /* Pass float64 interrupts */
        pasynManager->interruptStart(pPvt->float64InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynFloat64Interrupt *pfloat64Interrupt = pnode->drvPvt;
            channel = pfloat64Interrupt->addr;
            reason = pfloat64Interrupt->pasynUser->reason;
            startPoint = STpoint;                                                   /* re-gain teh start point */
            if (reason == hy8401Data) 
            {

                dValue = 0.0;
                for(i=0; i<numOfElement; i++)
                {
                    dValue += toVoltage(pPvt, (double) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel), channel);
                    startPoint--;
                    if(startPoint < 0) startPoint = BANKSIZE - 1;                   /* if pointer reaches the bottom of the memory, reset it */
                }
                if (numOfElement != 0)                                              /* safety check */
                    dValue = dValue / numOfElement;
                else
                    dValue = dValue; 
    
                pfloat64Interrupt->callback(pfloat64Interrupt->userPvt, 
                                            pfloat64Interrupt->pasynUser,
                                            dValue);
            } 

            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->float64InterruptPvt);

        /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/
        pointer = data[0];
        pPvt->lastPointer = data[0];                                                /* store pointer for periodic records (poll mode) */

        switch(pPvt->mode)
        {
            case hy8401Continuous:
                if (pointer >= pPvt->samples) 
                    startPoint = pointer - pPvt->samples;
                else
                    startPoint = BANKSIZE - (pPvt->samples - pointer);
                numOfElement = pPvt->samples;
                break;
            case hy8401Triggered:
                startPoint = pointer - pPvt->samples;
                numOfElement = pPvt->samples;
                break;
            case hy8401Gated:
                if (pPvt->offset >= 0)                                              /* offset is positive */
                {
                    if (pointer > pPvt->offset)                                     /* number of data collected is greater than offset */
                    {
                        startPoint = pPvt->offset;
                        numOfElement = pointer - pPvt->offset;
                    }else                                                           /* number of data collected is less than offset, no enough data, return all collected */
                    {
                        startPoint = 0;
                        numOfElement = pointer;
                    }
                }else                                                               /* offset is negative */
                {
                    if (pointer > (-pPvt->offset))
                    {
                        startPoint = pointer + pPvt->offset;
                        numOfElement = -pPvt->offset;                 
                    }else
                    {
                        startPoint = 0;
                        numOfElement = pointer;                 
                    }
                }
                break;
        }
        STpoint = startPoint;                                                       /* store start point */

        /* Pass int32Array interrupts */
        pasynManager->interruptStart(pPvt->int32ArrayInterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynInt32ArrayInterrupt *pint32ArrayInterrupt = pnode->drvPvt;
            reason = pint32ArrayInterrupt->pasynUser->reason;
            channel = pint32ArrayInterrupt->addr;
         	startPoint = STpoint;                                                   /* re-gain teh start point */
            if (reason == hy8401Data) 
            {

                for(i=0; i<numOfElement; i++)
                {
                    pPvt->ladcData[channel][i] = calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel);
                    startPoint++;
                    if(startPoint >= BANKSIZE) startPoint = 0;                      /* if pointer reaches the bottom of the memory, reset it */
                }
                pint32ArrayInterrupt->callback(pint32ArrayInterrupt->userPvt, 
                                               pint32ArrayInterrupt->pasynUser,
                                               (epicsInt32*) pPvt->ladcData[channel], 
                                               numOfElement);
            } 

            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->int32ArrayInterruptPvt);

        /* Pass float64Array interrupts */
        pasynManager->interruptStart(pPvt->float64ArrayInterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynFloat64ArrayInterrupt *pfloat64ArrayInterrupt = pnode->drvPvt;
            reason = pfloat64ArrayInterrupt->pasynUser->reason;
            channel = pfloat64ArrayInterrupt->addr;
        	startPoint = STpoint;                                                   /* re-gain teh start point */
            if (reason == hy8401Data) 
            {

                for(i=0; i<numOfElement; i++)
                {
                    pPvt->dadcData[channel][i] = toVoltage(pPvt, (double) calibrate(pPvt, (double) (pPvt->convMemoryBase + channel * BANKSIZE)[startPoint], channel), channel);
                    startPoint++;
                    if(startPoint >= BANKSIZE) startPoint = 0;                      /* if pointer reaches the bottom of the memory, reset it */
                }
                pfloat64ArrayInterrupt->callback(pfloat64ArrayInterrupt->userPvt, 
                                               pfloat64ArrayInterrupt->pasynUser,
                                               (epicsFloat64*) pPvt->dadcData[channel], 
                                               numOfElement);
            }
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->float64ArrayInterruptPvt);

        /* re-enable interrupt */
        switch (pPvt->mode)
        {
            case hy8401Continuous:                                                  /* continuous mode re-initialise the adc in isr. Here nothing needs to be done */
                break;
            case hy8401Triggered:
            case hy8401Gated:
                if (!pPvt->reEnableState)                                           /* if re-trigger is disabled, stop further triggers. To trigger again, need to do an enable command */
                    setHy8401EnableDisable(pPvt, DISABLEADC);                       /* disable trigger */
                else
                    setHy8401EnableDisable(pPvt, ENABLEADC);                        /* enable trigger */
                break;
            default:
                break;
        }
        ipmIrqCmd(pPvt->carrierNum, pPvt->ipSlotNum, 0, ipac_irqEnable);
    }
}

static void intSweepQueuedTask(drvHy8401Pvt *pPvt)
{
    int i;
    int reason;
    int sweepData[MAX_ADC_CHANNELS + 2];
    ELLLIST *pclientList;
    interruptNode *pnode;
    int channel;
    double dValue = 0.0;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(pPvt->intSweepQId, sweepData, sizeof(sweepData));

        /* Pass float64 interrupts */
        pasynManager->interruptStart(pPvt->float64InterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynFloat64Interrupt *pfloat64Interrupt = pnode->drvPvt;
            channel = pfloat64Interrupt->addr;
            reason = pfloat64Interrupt->pasynUser->reason;

            if (reason == hy8401SweepData)                                              /* this is for EPID records in EPID case */              
            {
                dValue = (double) calibrate(pPvt, (double) sweepData[channel + 2], channel);
                pfloat64Interrupt->callback(pfloat64Interrupt->userPvt, 
                                            pfloat64Interrupt->pasynUser,
                                            dValue);
            }

            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->float64InterruptPvt);

        /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/

        /* Pass int32Array interrupts */
        pasynManager->interruptStart(pPvt->int32ArrayInterruptPvt, &pclientList);
        pnode = (interruptNode *)ellFirst(pclientList);
        while (pnode) {

            asynInt32ArrayInterrupt *pint32ArrayInterrupt = pnode->drvPvt;
            reason = pint32ArrayInterrupt->pasynUser->reason;
            channel = pint32ArrayInterrupt->addr;
            if (reason == hy8401SweepData)                                              /* this is for fastsweep records in MCA case */              
            {
                for(i=0; i<MAX_ADC_CHANNELS; i++)
                    pPvt->chanData[i] = calibrate(pPvt, (double) sweepData[i + 2], i);

                pint32ArrayInterrupt->callback(pint32ArrayInterrupt->userPvt, 
                                               pint32ArrayInterrupt->pasynUser,
                                               (epicsInt32*) pPvt->chanData, 
                                               MAX_ADC_CHANNELS);                
            }
            pnode = (interruptNode *)ellNext(&pnode->node);
        }
        pasynManager->interruptEnd(pPvt->int32ArrayInterruptPvt);
        ipmIrqCmd(pPvt->carrierNum, pPvt->ipSlotNum, 0, ipac_irqEnable);

    }
}

/*********************************************************/
/*                                                      */
/* Calibration Routines                                */
/*                                                    */
/*****************************************************/

/* This routine linearly rescales an input in the range [inLow,inHigh]
 * into an output in the range [outLow,outHigh].  The result is clipped
 * to fit in the 32-bit range. */

static long Rescale(
    int inLow, int inHigh, long outLow, long outHigh, double in)
{
    double scale =
        ((double) outHigh - outLow) / ((double) inHigh - inLow);
    double result = (scale * (in - inLow) + outLow);

    /* Clip output to 32-bit range. */
    if (result <= LONG_MIN)
		return (long) (LONG_MIN / 65536);		/* changed by Peter Leicester for clamping the maximum to about +/-10v */
    else if (result >= LONG_MAX)
        return (long) (LONG_MAX / 65536);		/* changed by Peter Leicester for clamping the maximum to about +/-10v */
    else
        return (long) (result / 65536);
}

/* Applies the calibration algorithm appropriate to the card to a raw
 * value, translating a nominally 16-bit unsigned input into a signed
 * 32-bit output.
 *
 * The calibration type is read from the PROM.  We support three
 * calibration types: 0, 1 and 2.  Type 0 is pre-calibrated in hardware
 * (with trims), while types 1 and 2 use break-points to determine
 * piece-wise linear calibration. */

static long calibrate(drvHy8401Pvt *pPvt, double rawval, int channel)
{
    struct CALDATA * cal = &pPvt->cal[channel];
/*    return (unsigned long) rawval;*/
    switch (pPvt->calType)
    {
        /* Older cards are pre-calibrated (trimmed) in the hardware at
         * time of manufacture. No further calibration should be
         * necessary.  We treat unrecognised calibration types in the
         * same way. */
        case 0:
        default:
            /* Scale into the calibrated output range, but otherwise leave
             * the value unchanged. */
            return Rescale(RAW_MIN, RAW_MAX, CAL_MIN, CAL_MAX, rawval);

        /* Three point calibration: calibration points at -10V, 0V and
         * +10V.  We linearally interpolate separately in the two
         * ranges. */
        case 1:
            /* Separate scaling for positive and negative values. */
            if (rawval < cal->zero)
                return Rescale(cal->ngFS, cal->zero, CAL_MIN, 0, rawval);
            else
                return Rescale(cal->zero, cal->psFS, 0, CAL_MAX, rawval);

        /* Five point calibration gives calibration: -10V, -5V, 0V, +5V
         * and +10V.  This gives us four linear interpolations. */
        case 2:
            if (rawval < cal->ngHS)
                return Rescale(
                    cal->ngFS, cal->ngHS, CAL_MIN, CAL_MIN/2, rawval);
            else if (rawval < cal->zero)
                return Rescale(cal->ngHS, cal->zero, CAL_MIN/2, 0, rawval);
            else if (rawval < cal->psHS)
                return Rescale(cal->zero, cal->psHS, 0, CAL_MAX/2, rawval);
            else
                return Rescale(
                    cal->psHS, cal->psFS, CAL_MAX/2, CAL_MAX, rawval);
    }
}


/* this routine convert the raw adc value to voltage (-10V ~ +10V in the case of 8401) */
static double toVoltage( drvHy8401Pvt *pPvt, double rawval, int channel )
{
    struct CALDATA * cal = &pPvt->cal[channel];
    return (double)((double)10 - 20 * ((double) cal->psFS - cal->zero - rawval) / ((double) cal->psFS - cal->ngFS));
}


/* Read the card calibration data from PROM. */
static void getcalibration(	drvHy8401Pvt *pPvt, int verbose)
{
    int     i;
/*	USHORT *promcal;*/

    /* store and parse the calibration type */

    if (verbose)
        printf("Calibration type = %d\n", pPvt->calType);
   
	switch (pPvt->calType)
	{
			/* Pre-calibrated hardware */
	        case 0:
		        /* no calibration data is stored on cards of this type */
			    break;

	        /* Three point linear interpolation */
		    case 1:
			    /* for each ADC channel on the card */
	            for (i = 0; i < MAX_ADC_CHANNELS; i++)
				{
					struct CAL3PT *romcal = (struct CAL3PT *)(&pPvt->adc->cal.Cal3pt[i].ngFS);
	                struct CALDATA *cal = &pPvt->cal[i];
		            /* Read the channel's calibration data.  For three
			         * point calibration we don't use the half-scale
				     * values; they're initialised here for definiteness. */
					cal->ngFS = romcal->ngFS;
		            cal->ngHS = 0;
			        cal->zero = romcal->zero;
				    cal->psHS = 0;
					cal->psFS = romcal->psFS;

	                /* check if we have been asked to be more chatty */
		            if (verbose)
			            printf("Cal data %d - %04x %04x %04x\n",
				            i, cal->ngFS, cal->zero, cal->psFS);
	            }
		        break;

			/* Five point linear interpolation */
	        case 2:
		        /* for each ADC channel on the card */
			    for (i = 0; i < MAX_ADC_CHANNELS; i++)
				{
					struct CAL5PT *romcal = (struct CAL5PT *)(&pPvt->adc->cal.Cal5pt[i].ngFS);
	                struct CALDATA *cal = &pPvt->cal[i];
		            /* read the channel's calibration data */
			        cal->ngFS = romcal->ngFS;
				    cal->ngHS = romcal->ngHS;
					cal->zero = romcal->zero;
	                cal->psHS = romcal->psHS;
		            cal->psFS = romcal->psFS;

			        /* Report calibration settings as read. */
				    if (verbose)
					    printf("Cal data %d - %04x %04x %04x %04x %04x\n",
						    i, cal->ngFS, cal->ngHS, cal->zero,
							cal-> psHS, cal->psFS);
				}
				break;

	        /* unknown/unsupported calibration type */
		    default:
			    /* report unsupported calibration type */
				printf("getcalibration: unknown calibration type %d found\n",
					pPvt->calType);

	            /* treat card as pre-calibrated (i.e. do nothing) */
		        pPvt->calType = 0;
			    break;
		}
}

