/* This is the EPICS dependent code for the Hytec 8401 ADC driver.
 * By making this separate file for the EPICS dependent code the driver itself
 * only needs libCom from EPICS for OS-independence (i.e. it is then possible to create
 * module tests without needing to load the entire EPICS libraries!)
 */
 
#include <iocsh.h>
#include <drvSup.h>
#include <epicsExport.h>

#include "drvHy8401ipAsyn.h"

/* Code for iocsh registration */

/* Hy8401Init */
static const iocshArg Hy8401ipInitArg0 = {"Asyn port name", iocshArgString};
static const iocshArg Hy8401ipInitArg1 = {"vmeSlotNum", iocshArgInt};
static const iocshArg Hy8401ipInitArg2 = {"ipSlotNum", iocshArgInt};
static const iocshArg Hy8401ipInitArg3 = {"vectorNum", iocshArgInt};
static const iocshArg Hy8401ipInitArg4 = {"samples", iocshArgInt};
static const iocshArg Hy8401ipInitArg5 = {"average", iocshArgInt};
static const iocshArg Hy8401ipInitArg6 = {"offset", iocshArgInt};
static const iocshArg Hy8401ipInitArg7 = {"Scan mode", iocshArgInt};
static const iocshArg Hy8401ipInitArg8 = {"clockRate", iocshArgInt};
static const iocshArg Hy8401ipInitArg9 = {"External Clock", iocshArgInt};
static const iocshArg Hy8401ipInitArg10 = {"fastADC", iocshArgInt};

static const iocshArg * const Hy8401ipInitArgs[] =  {&Hy8401ipInitArg0,
                                                       &Hy8401ipInitArg1,
                                                       &Hy8401ipInitArg2,
                                                       &Hy8401ipInitArg3,
                                                       &Hy8401ipInitArg4,
                                                       &Hy8401ipInitArg5,
                                                       &Hy8401ipInitArg6,
                                                       &Hy8401ipInitArg7,
                                                       &Hy8401ipInitArg8,
                                                       &Hy8401ipInitArg9,
                                                       &Hy8401ipInitArg10};

static const iocshFuncDef initHy8401ip = {"Hy8401ipAsynInit", 12, Hy8401ipInitArgs};
static void initHy8401ipCallFunc(const iocshArgBuf *args)
{
    Hy8401ipAsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival, args[4].ival,
                   args[5].ival, args[6].ival, args[7].ival, args[8].ival, args[9].ival, args[10].ival);
}


/* Hy8401Config */


static const iocshArg Hy8401ipConfigArg0 = {"Asyn port name", iocshArgString};
static const iocshArg Hy8401ipConfigArg1 = {"aiType", iocshArgInt};
static const iocshArg Hy8401ipConfigArg2 = {"range", iocshArgInt};
static const iocshArg Hy8401ipConfigArg3 = {"firstChan", iocshArgInt};
static const iocshArg Hy8401ipConfigArg4 = {"lastChan", iocshArgInt};
static const iocshArg Hy8401ipConfigArg5 = {"chanBuffSize", iocshArgInt};

static const iocshArg * const Hy8401ipConfigArgs[] =  {&Hy8401ipConfigArg0,
                                                       &Hy8401ipConfigArg1,
                                                       &Hy8401ipConfigArg2,
                                                       &Hy8401ipConfigArg3,
                                                       &Hy8401ipConfigArg4,
                                                       &Hy8401ipConfigArg5};

static const iocshFuncDef configHy8401ip = {"Hy8401ipAsynConfig", 5, Hy8401ipConfigArgs};
static void configHy8401ipCallFunc(const iocshArgBuf *args)
{
    Hy8401ipAsynConfig(args[0].sval, args[1].ival, args[2].ival, args[3].ival, args[4].ival, args[5].ival);
}

static void Hy8401ipAsynRegister(void)
{
    iocshRegister(&initHy8401ip, initHy8401ipCallFunc);
    iocshRegister(&configHy8401ip, configHy8401ipCallFunc);
}

epicsExportRegistrar(Hy8401ipAsynRegister);


